from __future__ import annotations

from enum import StrEnum, auto
import time
from typing import Any, Callable, Iterable, Set

from openpectus.lang.exec.events import EventListener
from openpectus.lang.exec.units import convert_value_to_unit, is_supported_unit


# Represents tag API towards interpreter

class SystemTagName(StrEnum):
    BASE = "Base"
    RUN_COUNTER = "Run Counter"
    BLOCK = "Block"
    BLOCK_TIME = "Block Time"
    PROCESS_TIME = "Process Time"
    RUN_TIME = "Run Time"
    CLOCK = "Clock"
    SYSTEM_STATE = "System State"
    METHOD_STATUS = "Method Status"
    CONNECTION_STATUS = "Connection Status"
    RUN_ID = "Run Id"

    # these tags are only present if defined in uod.
    BLOCK_VOLUME = "Block Volume"
    ACCUMULATED_VOLUME = "Accumulated Volume"
    BLOCK_CV = "Block CV"
    ACCUMULATED_CV = "Accumulated CV"


    @staticmethod
    def has_value(value: str):
        """ Determine if enum has this string value defined. Case sensitive. """
        return value in SystemTagName.__members__.values()


TagValueType = int | float | str | None
""" Represents the types of tag values """

TagFormatFunction = Callable[[Any], str]
""" Represents a function that is used to format tag values for display """


def format_time_as_clock(value: float) -> str:
    import datetime
    date = datetime.datetime.fromtimestamp(value, datetime.UTC)
    tm = date.time()
    return f"{tm.hour:02}:{tm.minute:02}:{tm.second:02}"


class ChangeListener():
    """ Collects named changes. Used by engine to track tag changes """

    def __init__(self) -> None:
        self._changes: Set[str] = set()

    def notify_change(self, elm: str):
        self._changes.add(elm)

    def clear_changes(self):
        self._changes.clear()

    @property
    def changes(self) -> list[str]:
        return list(self._changes)


class ChangeSubject():
    """ Inherit to support change notification. Used by engine to track tag changes """

    def __init__(self) -> None:
        super().__init__()

        self._listeners: list[ChangeListener] = []

    def add_listener(self, listener: ChangeListener):
        self._listeners.append(listener)

    def notify_listeners(self, elm: str):
        for listener in self._listeners:
            listener.notify_change(elm)

    def clear_listeners(self):
        # for listener in self._listeners:
        #     listener.clear_changes()
        self._listeners.clear()


# seems only relevant in hw interface
class TagDirection(StrEnum):
    """ Specifies whether a tag is read from or written to hardware and whether is can be changed in UI.

    Direction of the tag is in relation to the physical IO. Sensors are regarded as inputs and
    actuators as outputs. Derived values are regarded as NA.
    """
    Input = auto()
    """ Tag is read from hardware, e.g. a sensor """
    Output = auto()
    """ Tag is written to hardware, e.g. an actuator """
    NA = auto()
    """ Tag is calculated/derived and neither read from or written to hw. """
    Unspecified = auto()


class Unset():
    """ Used to specify that a value has not been set.

    Used for nullable values to distinguish between being set to None and not being set.
    """
    pass


class Tag(ChangeSubject, EventListener):
    """ Base class for tags. Most tags do not need their own class but can just use this class.

    Supports change tracking which is used by engine to detect changes between reads of hardware values.

    Supports lifetime notification events that are automatically invoked by the engine.
    """
    def __init__(
            self,
            name: str,
            tick_time: float | None = None,
            value: TagValueType = None,
            unit: str | None = None,
            direction: TagDirection = TagDirection.NA,
            safe_value: TagValueType | Unset = Unset(),
            format_fn: TagFormatFunction | None = None
            ) -> None:

        super().__init__()

        assert name is not None
        assert name != ""
        assert isinstance(name, str)  # this includes StrEnum which is useful

        if unit is not None:
            if not isinstance(unit, str):
                raise ValueError("unit must be None or a string")
            elif not is_supported_unit(unit):
                raise ValueError(f"Invalid unit '{unit}'")

        self.name: str = name
        self.tick_time = tick_time or time.time()
        self.value: TagValueType = value  # Do we need default also? sometimes it is used as safe but are the other uses?
        self.unit: str | None = unit
        self.choices: list[str] | None = None
        self.direction: TagDirection = direction
        self.safe_value: TagValueType | Unset = safe_value
        self.format_fn = format_fn

    def as_readonly(self) -> TagValue:
        """ Convert the value to a readonly and immutable TagValue instance """
        value_formatted = None if self.format_fn is None else self.format_fn(self.get_value())
        return TagValue(self.name, self.tick_time, self.value, value_formatted, self.unit, self.direction)

    def set_value(self, val: TagValueType, tick_time: float) -> None:
        if val != self.value:
            self.value = val
            self.tick_time = tick_time
            self.notify_listeners(self.name)

    def set_value_and_unit(self, val: TagValueType, unit: str, tick_time: float) -> None:
        """ Set a new value by converting the provided value and unit into the the unit of the tag. """
        if not isinstance(val, (int, float,)):
            raise ValueError(f"Cannot set unit for a non-numeric value {val} of type {type(val).__name__}")
        if self.unit is None:
            raise ValueError("Cannot change unit on a tag with no unit")
        val = convert_value_to_unit(val, unit, self.unit)
        self.set_value(val, tick_time)

    def get_value(self):
        return self.value

    def as_number(self) -> int | float:
        if not isinstance(self.value, (int, float)):
            raise ValueError(
                f"Value is not numerical: '{self.value}' has type '{type(self. value).__name__}' tag: '{self.name}'")
        return self.value

    def as_float(self) -> float:
        if not isinstance(self.value, (float,)):
            raise ValueError(
                f"Value is not a float: '{self.value}' has type '{type(self. value).__name__}' tag: '{self.name}'")
        return self.value

    def archive(self) -> str | None:
        """ The value to write to archive or None to skip that tag from archival """
        if self.value is None:
            return ""
        elif isinstance(self.value, float):
            return f"{self.as_float():0.5f}"
        else:
            return str(self.value)


class TagCollection(ChangeSubject, ChangeListener, Iterable[Tag]):
    """ Represents a  name/tag dictionary. """
    def __init__(self, tags: Iterable[Tag] | None = None) -> None:
        super().__init__()
        self.tags: dict[str, Tag] = {}
        if tags is not None:
            for tag in tags:
                self.add(tag, exist_ok=False)

    def as_readonly(self) -> TagValueCollection:
        return TagValueCollection([t.as_readonly() for t in self.tags.values()])

    # propagate tag changes to collection changes
    def notify_change(self, elm: str):
        self.notify_listeners(elm)

    @property
    def names(self) -> list[str]:
        """ Return the tag names """
        return list(self.tags.keys())

    def __iter__(self):
        yield from self.tags.values()

    def __getitem__(self, tag_name: str) -> Tag:
        return self.tags[tag_name]

    def __len__(self) -> int:
        return len(self.tags)

    def get(self, tag_name: str) -> Tag:
        if tag_name is None or tag_name.strip() == '':
            raise ValueError("tag_name is None or empty")
        if tag_name not in self.tags.keys():
            raise ValueError(f"Tag name {tag_name} not found")
        return self[tag_name]

    def add(self, tag: Tag, exist_ok: bool = True):
        """ Add tag to collection. If tag name already exists and exist_ok is False, a ValueError is raised. """
        if tag is None:
            raise ValueError("tag is None")
        if tag.name is None or tag.name.strip() == '':
            raise ValueError("tag name is None or empty")
        if tag.name in self.tags.keys() and not exist_ok:
            raise ValueError(f"A tag named {tag.name} already exists")

        self.tags[tag.name] = tag
        tag.add_listener(self)

    def with_tag(self, tag: Tag):
        self.add(tag)
        return self

    def has(self, tag_name: str) -> bool:
        if tag_name is None or tag_name.strip() == '':
            raise ValueError("tag_name is None or empty")
        return tag_name in self.tags.keys()

    def get_value_or_default(self, tag_name) -> TagValueType:
        if not self.has(tag_name):
            return None
        return self.tags[tag_name].get_value()

    def merge_with(self, other: TagCollection) -> TagCollection:
        """ Returns a new TagCollection with the combined tags of both collections.

        Tag instances are kept so tag values are shared between the three collections.

        In case of duplicate tag names, tags from other collection are used.
        """
        tags = TagCollection()
        for tag in self.tags.values():
            tags.add(tag)
        for tag in other.tags.values():
            tags.add(tag)
        return tags

    @staticmethod
    def create_system_tags() -> TagCollection:
        tags = TagCollection([
            Tag(SystemTagName.BASE, value="min"),  # note special value "min" and no unit
            Tag(SystemTagName.RUN_COUNTER, value=0),
            Tag(SystemTagName.BLOCK, value=None),
            Tag(SystemTagName.BLOCK_TIME, value=0.0, unit="s", format_fn=format_time_as_clock),
            Tag(SystemTagName.PROCESS_TIME, value=0.0, unit="s", format_fn=format_time_as_clock),
            Tag(SystemTagName.RUN_TIME, value=0.0, unit="s", format_fn=format_time_as_clock),
            Tag(SystemTagName.CLOCK, value=0.0, unit="s", format_fn=format_time_as_clock),
            Tag(SystemTagName.SYSTEM_STATE, value="Stopped"),
            Tag(SystemTagName.METHOD_STATUS, value="OK"),
            Tag(SystemTagName.CONNECTION_STATUS, value="Disconnected"),
            Tag(SystemTagName.RUN_ID, value=None),
        ])
        return tags


class TagValue():
    """ Read-only and immutable representation of a tag value. """
    def __init__(
            self,
            name: str,
            tick_time: float = 0.0,
            value: TagValueType = None,
            value_formatted: str | None = None,
            unit: str | None = None,
            direction: TagDirection = TagDirection.Unspecified,
    ):
        if name is None or name.strip() == '':
            raise ValueError("name is None or empty")

        self.name = name
        self.tick_time = tick_time
        self.value = value
        self.value_formatted = value_formatted
        self.unit = unit
        self.direction = direction


class TagValueCollection(Iterable[TagValue]):
    """ Represents a read-only and immutable dictionary of tag values. """

    def __init__(self, values: Iterable[TagValue]) -> None:
        super().__init__()
        self._tag_values: dict[str, TagValue] = {}
        for v in values:
            self._add(v)

    @staticmethod
    def empty() -> TagValueCollection:
        return TagValueCollection([])

    def get(self, tag_name: str) -> TagValue:
        if tag_name is None or tag_name.strip() == '':
            raise ValueError("tag_name is None or empty")
        if tag_name not in self._tag_values.keys():
            raise ValueError(f"Tag name {tag_name} not found")
        return self[tag_name]

    def has(self, tag_name: str) -> bool:
        if tag_name is None or tag_name.strip() == '':
            raise ValueError("tag_name is None or empty")
        return tag_name in self._tag_values.keys()

    def _add(self, tag_value: TagValue):
        self._tag_values[tag_value.name] = tag_value

    def __iter__(self):
        yield from self._tag_values.values()

    def __getitem__(self, tag_name: str) -> TagValue:
        return self._tag_values[tag_name]

    def __len__(self) -> int:
        return len(self._tag_values)

    def to_list(self) -> list[TagValue]:
        return [v for v in self._tag_values.values()]



from __future__ import annotations
from typing import Callable, List
from uuid import UUID, uuid4


class PNode():
    """ Represents an AST node in an openpectus runtime model object tree. """

    def __init__(self, parent: PNode | None) -> None:
        """ Create new node with specified parent or None for the root node. If parent is given,
            the created node is added as child of the parent.
        """
        self.parent = parent
        """ The parent node or None for the root node. """

        self.children: List[PNode] | None = None
        """ The child nodes of the node, if any. None if the node cannot have child nodes. """

        self._id: UUID = uuid4()
        """ Node id, assigned randomly during initialization. Can be changed for the sole
        purpose of merging two parse trees. """

        self.is_root: bool = False
        """ Set to true for the root node. False for all other nodes in a tree. """

        self.indent: int | None = None
        """ The indentation for the line, expressed in number of spaces. A multiple of
        pprogrambuilder.INDENTATION_SPACES. """

        self.line: int | None = None
        """ The line number, starting from 1. All instructions have unique line numbers,
        except PProgram which shares line number with its first child instruction.
        """

        self.errors: List[PError] | None = None
        """ Errors encountered during parsing. """

        self._inst_error: PInstError | None = None
        """ Partial parse error. Is set by parser. Must be processed into one or more elements in self.errors
        for general consumption
         """

        self._cancellable: bool = False
        """ Whether node is statically cancellable, i.e. supports the cancel operation at all. """

        self._cancelled: bool = False
        self._forcible: bool = False
        """ Whether node is statically forcible, i.e. supports the force operation at all. """

        self._forced: bool = False

        if self.parent is not None and self.parent.children is not None:
            self.parent.children.append(self)

    @property
    def id(self) -> UUID:
        return self._id

    def update_id(self, new_id: UUID):
        self._id = new_id

    @property
    def display_name(self) -> str:
        """ The name of the instruction for general display"""
        if hasattr(self, "name"):
            return getattr(self, "name")
        return str(self)

    @property
    def runlog_name(self) -> str | None:
        """ Name to show in runlog. Return None to skip in runlog. """
        return None

    @property
    def instruction_name(self) -> str | None:
        """ Name to query for in tests. Return None to not support querying.

        Should not include arguments because their formatting may not be predictable. """
        return None

    @property
    def depth(self):
        if self.parent is None:
            return 0
        return self.parent.depth + 1

    @property
    def cancellable(self) -> bool:
        """ Whether the node is cancellable in its current state. Virtual property. """
        return self._cancellable and not self._cancelled and not self._forced

    @property
    def cancelled(self) -> bool:
        """ Whether the node has been cancelled. Virtual property. """
        return self._cancelled

    def cancel(self):
        if self.cancellable:
            self._cancelled = True

    @property
    def forcible(self) -> bool:
        """ Whether the node is forcible in its current state. Virtual property. """
        return self._forcible and not self._forced and not self._cancelled

    @property
    def forced(self) -> bool:
        """ Whether the node has been forced. Virtual property. """
        return self._forced

    def force(self):
        if self.forcible:
            self._forced = True

    def __str__(self) -> str:
        return type(self).__name__

    def __eq__(self, other) -> bool:
        # TODO: Here be dragons. See issue https://github.com/Open-Pectus/Open-Pectus/issues/22 regarding
        # identity vs value equality, performance and caching.
        # See also note in next_following() below.
        # Current implementation may cause RecursionError
        if not isinstance(other, PNode):
            return False
        return self.id == other.id
        # if self.depth != other.depth:
        #     return False
        # return self.__dict__ == other.__dict__

    def get_child_nodes(self, recursive: bool = False) -> List[PNode]:
        children: List[PNode] = []
        if self.children is None:
            return children
        for child in self.children:
            children.append(child)
            if recursive:
                children.extend(child.get_child_nodes(recursive))
        return children

    def add_error(self, error: PError | str):
        if self.errors is None:
            self.errors = []
        if isinstance(error , PError):
            self.errors.append(error)
        else:
            self.errors.append(PError(error))

    def has_error(self, recursive: bool = False):
        if not recursive:
            if self.errors is None or len(self.errors) == 0:
                return False
            return True
        else:
            for c in self.get_child_nodes(True):
                if c.has_error():
                    return True
            return False

    def collect_errors(self):
        """ Populate the errors collection from node state that indicates parse errors. """
        if self._inst_error is not None:
            self.add_error("Syntax error")
        for child in self.get_child_nodes(recursive=True):
            child.collect_errors()

    def iterate(self, fn: Callable[[PNode], None]):
        fn(self)
        if self.children is not None:
            for child in self.children:
                child.iterate(fn)

    def has_ancestor(self, ancestor_candidate: PNode):
        parent = self.parent
        while parent is not None:
            if parent == ancestor_candidate:
                return True
            parent = parent.parent
        return False

    def reset_runtime_state(self, recursive: bool = False):
        """ Override to clear node runtime state. """
        if recursive:
            def reset(node: PNode):
                if node != self:
                    node.reset_runtime_state(False)
            self.iterate(reset)


class PProgram(PNode):
    """ Represents a complete pectus program. """
    def __init__(self) -> None:
        super().__init__(None)
        self.children = []
        self.is_root = True
        self.line = 1

    def get_instructions(self, include_blanks: bool = False) -> List[PInstruction]:
        """ Return list of all program instructions, recursively, depth first. """
        children: List[PInstruction] = self.get_child_nodes(recursive=True)  # type: ignore
        if not include_blanks:
            return [n for n in children if not isinstance(n, PBlank)]
        return children

    def get_commands(self) -> List[PCommand]:
        return [c for c in self.get_instructions() if isinstance(c, PCommand)]

    def get_condition_nodes(self) -> List[PAlarm | PWatch]:
        return [c for c in self.get_instructions() if isinstance(c, PAlarm) or isinstance(c, PWatch)]


class PInjectedNode(PNode):
    """ Represents a subtree of injected code """
    def __init__(self, parent: PNode | None) -> None:
        super().__init__(parent)
        self.children = []


class PInstruction(PNode):
    """ Represents a generic instruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.time: float | None = None
        """ The delay threshold specified for the instruction measured in the Base unit. """

        self.comment: str = ''
        self._forcible = True

    @property
    def forcible(self) -> bool:
        return super().forcible and self.time is not None

    def reset_runtime_state(self, recursive: bool = False):
        self._forced = False
        super().reset_runtime_state(recursive=recursive)


class PBlock(PInstruction):
    """ Represents a Block intruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.children = []
        self.name: str = ''

    def __str__(self) -> str:
        return super().__str__() + ": " + self.name

    @property
    def runlog_name(self) -> str | None:
        return "Block: " + self.name

    @property
    def instruction_name(self) -> str | None:
        return "Block"


class PEndBlock(PInstruction):
    """ Represents an End block intruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

    @property
    def runlog_name(self) -> str | None:
        return "End block"

    @property
    def instruction_name(self) -> str | None:
        return "End block"


class PEndBlocks(PInstruction):
    """ Represents an End blocks intruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

    @property
    def runlog_name(self) -> str | None:
        return "End blocks"

    @property
    def instruction_name(self) -> str | None:
        return "End blocks"


class PWatch(PInstruction):
    """ Represents a Watch instruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.children = []
        self.condition: PCondition | None = None
        self.activated: bool = False

    # override cancellable and forcible to be disabled once activated
    @property
    def cancellable(self) -> bool:
        return not self.cancelled and not self.forced and not self.activated

    @property
    def forcible(self) -> bool:
        return not self.cancelled and not self.forced and not self.activated

    @property
    def condition_str(self) -> str:
        return self.condition.condition_str if self.condition is not None else ""

    def __str__(self) -> str:
        return super().__str__() + f": {self.condition_str} | line: {self.line}"

    @property
    def runlog_name(self) -> str | None:
        return "Watch: " + self.condition_str

    @property
    def instruction_name(self) -> str | None:
        return "Watch"

    def collect_errors(self):
        super().collect_errors()
        if self.condition is None:
            self.add_error(PError("Missing condition"))

    def reset_runtime_state(self, recursive: bool = False):
        self.activated = False
        self._cancelled = False
        self._forced = False
        super().reset_runtime_state(recursive=recursive)


class PAlarm(PInstruction):
    """ Represents an Alarm instruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.children = []
        self.condition: PCondition | None = None
        self.activated: bool = False

    # override cancellable and forcible to be disabled once activated
    @property
    def cancellable(self) -> bool:
        return not self.cancelled and not self.activated

    @property
    def forcible(self) -> bool:
        return not self.forced and not self.activated

    @property
    def condition_str(self) -> str:
        return self.condition.condition_str if self.condition is not None else ""

    def __str__(self) -> str:
        cnd = self.condition.condition_str if self.condition is not None else ""
        return super().__str__() + f": {cnd} | line: {self.line}"

    @property
    def runlog_name(self) -> str | None:
        return "Alarm: " + self.condition_str

    @property
    def instruction_name(self) -> str | None:
        return "Alarm"

    def collect_errors(self):
        super().collect_errors()
        if self.condition is None:
            self.add_error(PError("Missing condition"))

    def reset_runtime_state(self, recursive: bool = False):
        self.activated = False
        self._cancelled = False
        self._forced = False
        super().reset_runtime_state(recursive=recursive)


class PMark(PInstruction):
    """ Represents an Mark instruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.children = []
        self.name: str = ''
        self._forcible = False

    def __str__(self) -> str:
        return super().__str__() + ": " + self.name

    @property
    def runlog_name(self) -> str | None:
        return "Mark: " + self.name

    @property
    def instruction_name(self) -> str | None:
        return "Mark"

class PCommand(PInstruction):
    """ Represents a Command instruction (Start, Stop, Restart, ...) as well as uod commands.

    Note: The cancellable property is not set for uod commands because the node does not know
    about the command. Runlog must set this property.
    """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.name: str = ''
        self.args: str = ''

    def __str__(self) -> str:
        return super().__str__() + ": " + self.name

    def cancel(self):
        # skip the cancellable check here as noted above
        self._cancelled = True

    @property
    def runlog_name(self) -> str | None:
        return self.name

    @property
    def instruction_name(self) -> str | None:
        return self.name

class PCommandWithDuration(PInstruction):
    """ Represents a Command instruction with a possible duration (Pause, Hold and Wait). """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)

        self.name: str = ''
        """ Command name """
        self.duration: PDuration | None = None
        """ Unresolved duration expression """

    @property
    def cancellable(self) -> bool:
        return False

    @property
    def forcible(self) -> bool:
        if self.name == "Wait" and not self._forced:
            return True
        return False

    def __str__(self) -> str:
        duration = "" if self.duration is None else ", duration: " + str(self.duration)
        return super().__str__() + ": " + self.name + duration

    @property
    def runlog_name(self) -> str | None:
        runlog_argument = "" if self.duration is None else ": " + self.duration.runlog_argument
        return self.name + runlog_argument

    @property
    def instruction_name(self) -> str | None:
        return self.name

    def collect_errors(self):
        super().collect_errors()
        if self.name == "Pause":
            if self.duration is not None and self.duration.error:
                self.add_error(PError("Invalid Pause arguments. Specify either no duration arguments or both time and unit"))
        elif self.name == "Hold":
            if self.duration is not None and self.duration.error:
                self.add_error(PError("Invalid Hold arguments. Specify either no duration arguments or both time and unit"))
        elif self.name == "Wait":
            if self.duration is None or self.duration.error:
                self.add_error(PError("Wait requires a time and unit"))


class PBlank(PInstruction):
    """ Represents an all-whitespace pcode line. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)


class PComment(PInstruction):
    """ Represents a comment instruction. """
    def __init__(self, parent: PNode) -> None:
        super().__init__(parent)


class PErrorInstruction(PInstruction):
    """ Represents a non-parsable instruction pcode line. """
    def __init__(self, parent: PNode, code: str) -> None:
        super().__init__(parent)
        self.children = []
        if self.errors is None or len(self.errors) == 0:
            self.errors = [PError("Parse error")]
        self.code: str = code

    def has_error(self, recursive: bool = False):
        return True

    def __str__(self) -> str:
        return super().__str__() + ": Code: " + self.code


class PInstError:
    """ Represents the error part of a parsed instruction. This differs from PErrorInstruction by being
    partly parsed. This allows better language assistance than PErrorInstruction. """
    def __init__(self, message: str | None = None) -> None:
        self.message: str | None = message


# --- Non-nodes ---

class PError:
    """ Represents an error in an instruction. """
    def __init__(self, message: str | None = None) -> None:
        self.message: str | None = message


class PCondition:
    """ Represents a condition expression for Alarm and Watch instructions.

    The condition is resolved by ConditionEnrichAnalyzer.
    """
    def __init__(self, condition_str: str) -> None:
        self.condition_str = condition_str
        """ Original condition string expression """
        self.op = ""
        """ Unresolved condition operator """
        self.lhs = ""
        """ Unresolved condition left-hand-side expression """
        self.rhs = ""
        """ Unresolved condition right-hand-side expression """

        # Note: error is True by default and only modified by the analyzer
        self.error : bool = True
        self.tag_name: str | None = None
        self.tag_value: str | None = None
        self.tag_unit: str | None = None
        self.tag_value_numeric: int | float | None = None


class PDuration:
    """ Represents a duration for PCommandWithDuration (Pause, Hold and Wait) instructions.

    The duration string is resolved by DurationEnrichAnalyzer.
    """

    def __init__(self, duration_str: str) -> None:
        self.duration_str: str = duration_str

        # Note: error is True by default and only modified by the analyzer
        self.error : bool = True
        self.time: float | None = None
        self.unit: str | None = None

    def __str__(self) -> str:
        return f"Duration {self.time} {self.unit}"

    @property
    def runlog_argument(self):
        if self.error:
            return "(error)"
        else:
            if self.time is None:
                return ""
            elif self.unit is None:
                return f"{self.time}"
            else:
                return f"{self.time}{self.unit}"



"""Add uod fields

Revision ID: 774e4df9ce40
Revises: 05c4bf88d0bb
Create Date: 2024-07-10 10:58:53.684832

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '774e4df9ce40'
down_revision: Union[str, None] = '05c4bf88d0bb'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.add_column(sa.Column('uod_filename', sa.String(), nullable=False, server_default=''))
        batch_op.add_column(sa.Column('uod_author_name', sa.String(), nullable=False, server_default=''))
        batch_op.add_column(sa.Column('uod_author_email', sa.String(), nullable=False, server_default=''))

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.drop_column('uod_author_email')
        batch_op.drop_column('uod_author_name')
        batch_op.drop_column('uod_filename')

    # ### end Alembic commands ###



from __future__ import annotations

import logging
import socket
from typing import Callable, Any, Awaitable
import json
import ssl
import platform

from fastapi_websocket_rpc import RpcMethodsBase, WebSocketRpcClient
from fastapi_websocket_rpc.rpc_methods import RpcResponse
from websockets.exceptions import ConnectionClosedError
import httpx
import tenacity

import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
from openpectus.protocol.exceptions import ProtocolException, ProtocolNetworkException
import openpectus.protocol.messages as M
from openpectus import __version__, __name__ as client_name
from openpectus.protocol.dispatch_interface import (
    AGGREGATOR_REST_PATH, AGGREGATOR_RPC_WS_PATH, AGGREGATOR_HEALTH_PATH, AGGREGATOR_AUTH_CONFIG_PATH
)
from openpectus.protocol.serialization import serialize, deserialize
from openpectus.aggregator.routers.dto import AuthConfig

logger = logging.getLogger(__name__)

EngineMessageHandler = Callable[[AM.AggregatorMessage], Awaitable[M.MessageBase]]
""" Handler in engine that handles aggregator messages of a given type """

engine_headers = {"User-Agent": f"{client_name}/{__version__}"}

if platform.system() == "Windows":
    # Use system certificates on Windows instead
    # of Anaconda provided certificates.
    ssl_context = ssl.create_default_context()
    ssl_context.load_default_certs()
else:
    # SSL library method load_default_certs only does
    # something useful on Windows. On other systems
    # the Anaconda certificate bundle will have to do.
    ssl_context = httpx.create_ssl_context()


class EngineDispatcher():
    """
    Engine dispatcher for the Aggregator-Engine Protocol.
    Allows sending messages via HTTP POST and JSON-RPC.
    """

    class EngineRpcMethods(RpcMethodsBase):
        def __init__(self, dispatcher: EngineDispatcher, engine_id: str):
            super().__init__()
            self.disp = dispatcher
            self.engine_id = engine_id

        # Note: this method cannot have a return type specified - this causes a RPC Reader task failed
        # which seems like a bug in pydantic
        # It should return a serialized M.MessageBase message as type str
        async def dispatch_message_async(self, message_json: dict[str, Any]):  # -> str:
            """ Handle RPC call from aggregator. Dispatch message to registered handler. """
            try:
                message = deserialize(message_json)
                assert isinstance(message, M.MessageBase)
            except Exception:
                msg = "Dispatch failed. Message deserialization failed."
                logger.error(msg, exc_info=True)
                return json.dumps(serialize(M.ErrorMessage(message=msg)))

            result = await self.disp.dispatch_message_async(message)
            logger.debug(f"Return type of dispatched message {type(message).__name__} was: {type(result).__name__}")
            return json.dumps(serialize(result))

        async def get_engine_id_async(self):
            # Note: Must be declared async to be usable by RPC
            return self.engine_id

    def __init__(self, aggregator_host: str, secure: bool, uod_options: dict[str, str], secret: str = "") -> None:
        super().__init__()
        self._aggregator_host = aggregator_host
        self._uod_name = uod_options.pop("uod_name")
        self._uod_author_name = uod_options.pop("uod_author_name")
        self._uod_author_email = uod_options.pop("uod_author_email")
        self._uod_filename = uod_options.pop("uod_filename")
        self._location = uod_options.pop("location")
        self._rpc_client: WebSocketRpcClient | None = None
        self._handlers: dict[type, Callable[[Any], Awaitable[M.MessageBase]]] = {}
        self._engine_id = None
        self._sequence_number = 1
        self._secret = secret

        http_scheme = "https" if secure else "http"
        ws_scheme = "wss" if secure else "ws"
        self._ssl_context: ssl.SSLContext | bool = ssl_context if secure else False
        self._post_url = f"{http_scheme}://{aggregator_host}{AGGREGATOR_REST_PATH}"
        self._health_url = f"{http_scheme}://{self._aggregator_host}{AGGREGATOR_HEALTH_PATH}"
        self._rpc_url = f"{ws_scheme}://{self._aggregator_host}{AGGREGATOR_RPC_WS_PATH}"
        self._auth_config_url = f"{http_scheme}://{self._aggregator_host}{AGGREGATOR_AUTH_CONFIG_PATH}"
        logger.info(f"Connecting to aggregator using urls:\n{self._post_url}\n{self._rpc_url}")

    def check_aggregator_alive(self) -> bool:
        try:
            resp = httpx.get(self._health_url, headers=engine_headers, verify=self._ssl_context)
        except httpx.ConnectError as ex:
            logger.error(f"Connection to Aggregator health end point {self._health_url} failed.")
            logger.info("Connection to Aggregator health end point failed.")
            logger.info(f"Status url: {self._health_url}")
            logger.info(f"Error: {ex}")
            return False

        if resp.is_error:
            logger.error(
                f"Aggregator health end point {self._health_url} returned an unsuccessful result: {resp.status_code}.",
                exc_info=True)
            logger.info("Aggregator health end point returned an unsuccessful result.")
            logger.info(f"Status url: {self._health_url}")
            logger.info(f"HTTP status code returned: {resp.status_code}")
            return False

        logger.debug(f"Aggregator health url {self._health_url} responded succesfully")
        return True

    def is_aggregator_authentication_enabled(self) -> bool:
        response = httpx.get(self._auth_config_url, headers=engine_headers, verify=self._ssl_context)
        auth_config = AuthConfig(**response.json())
        return auth_config.use_auth

    async def connect_async(self):
        def logerror(retry_state: tenacity.RetryCallState):
            if retry_state and retry_state.outcome:
                ex = retry_state.outcome.exception()
                if ex:
                    logger.exception(ex)
        retry_config = {  # use no retry because we use our own reconnect mechanism
            'wait': tenacity.wait.wait_none,
            'retry': tenacity.retry_never,
            'reraise': True,
            "retry_error_callback": logerror
        }

        # invoke registration rest call that provides engine_id
        if self._engine_id is None:
            self._engine_id = await self._register_for_engine_id_async()
            if self._engine_id is None:
                logger.error("Failed to register because Aggregator refused the registration.")
                raise ProtocolNetworkException("Registration failed")

        rpc_methods = EngineDispatcher.EngineRpcMethods(self, self._engine_id)
        self._rpc_client = WebSocketRpcClient(
            uri=self._rpc_url,
            methods=rpc_methods,
            retry_config=retry_config,
            on_disconnect=[self.on_disconnect],
            user_agent_header=engine_headers,
            ssl=self._ssl_context if self._ssl_context else None,
        )
        try:
            await self._rpc_client.__aenter__()
        except Exception:
            raise ProtocolNetworkException("Error creating websocket connection")

        logger.info("Websocket connected")

    async def on_disconnect(self, channel):
        # Hooking up this handler with an INFO message seems to mostly avoid the reconnect
        # hang issue that sometimes occur when aggregator is down very briefly. Not all the
        # issues though. The offending call is WebSocketRpcClient.close() which logs this message:
        # INFO:fastapi_ws_rpc.RPC_CLIENT:Closing RPC client
        logger.info("on_websocket_disconnect")
        # asyncio.create_task(self.disconnect_async()) # this does not help - it must be one of the
        # calls in WebSocketRpcClient.close()

    async def disconnect_async(self):
        logger.info("Websocket disconnected")
        if self._rpc_client is not None:
            await self._rpc_client.__aexit__()
            self._rpc_client = None

    async def send_registration_msg_async(self, message: EM.RegisterEngineMsg) -> M.MessageBase:
        """ Send message registration message via post"""
        try:
            message_json = serialize(message)
        except Exception:
            logger.error("Message serialization failed", exc_info=True)
            return M.ErrorMessage(message="Message serialization failed")

        try:
            async with httpx.AsyncClient(verify=self._ssl_context) as client:
                response = await client.post(url=self._post_url, json=message_json, headers=engine_headers)
        except Exception as ex:
            logger.error(f"Post failed with  exception type {type(ex).__name__}")  # skip details,  exc_info=True)
            raise ProtocolNetworkException("Post failed with exception")

        if response.status_code == 200:
            response_json = response.json()
            try:
                value = deserialize(response_json)
                return value
            except Exception:
                logger.error("Message deserialization failed", exc_info=True)
                return M.ErrorMessage(message="Post succeeded but result deserialization failed")
        else:
            message_type = type(message)
            logger.error(f"Non-success http status code: {response.status_code} for input message type: {message_type}")
            raise ProtocolNetworkException(f"Post failed with non-success http status code: {response.status_code}" +
                                           " for input message type: {message_type}")

    async def send_async(self, message: EM.EngineMessage) -> M.MessageBase:
        """ Send message via websocket. """
        assert self._rpc_client is not None, "Cannot send when rpc_client is None"

        if self._engine_id is None:
            # this should never occur since engine_id is required to set up the websocket connection
            raise ProtocolException("Engine did not have engine_id yet")

        message.engine_id = self._engine_id
        self.assign_sequence_number(message)

        try:
            message_json = serialize(message)
        except Exception:
            logger.error("Message serialization failed", exc_info=True)
            return M.ErrorMessage(message="Message serialization failed")

        try:
            response = await self._rpc_client.other.dispatch_message_async(message_json=message_json)
            assert isinstance(response, RpcResponse)
            logger.debug(f"Sent message: {message.ident}")
        except ConnectionClosedError:
            raise ProtocolNetworkException("Connection closed")
        except Exception:
            logger.error(f"Unkown error sending message: {message.ident}", exc_info=True)
            return M.ErrorMessage(message=f"Unkown error sending message: {message.ident}")

        value_str = response.result
        try:
            assert isinstance(value_str, str)
            value = deserialize(json.loads(value_str))
            assert isinstance(value, M.MessageBase), f"Expected type M.MessageBase, not {type(value).__name__}"
            return value
        except Exception:
            err = f"Error deserializing response for message: {message.ident}"
            logger.error(err, exc_info=True)
            return M.ErrorMessage(message=err)


    def set_rpc_handler(self, message_type: type[AM.AggregatorMessage], handler: EngineMessageHandler):
        """ Register handler for given message_type. """
        if message_type in self._handlers.keys():
            logger.error(f"Handler for message type {message_type} is already set.")
        self._handlers[message_type] = handler

    async def dispatch_message_async(self, message: M.MessageBase) -> M.MessageBase:
        """ Dispatch incoming message to registered handler. """
        message_type = type(message)
        try:
            result = await self._handlers[message_type](message)
            assert isinstance(result, M.MessageBase)
            return result
        except KeyError:
            msg = f"Dispatch failed for message type: {message_type}. No handler registered."
            logger.warning(msg)
            return M.ErrorMessage(message=msg)
        except Exception:
            msg = f"Dispatch failed for message type: {message_type}. A handler was registered, but failed somehow."
            logger.warning(msg, exc_info=True)
            return M.ErrorMessage(message=msg)

    async def _register_for_engine_id_async(self) -> str | None:
        logger.info("Registering for engine id")
        register_engine_msg = EM.RegisterEngineMsg(
            computer_name=socket.gethostname(),
            uod_name=self._uod_name,
            uod_author_name=self._uod_author_name,
            uod_author_email=self._uod_author_email,
            uod_filename=self._uod_filename,
            location=self._location,
            engine_version=__version__,
            secret=self._secret)
        register_response = await self.send_registration_msg_async(register_engine_msg)
        if not isinstance(register_response, AM.RegisterEngineReplyMsg) or not register_response.success:
            logger.warning("Aggregator refused registration")
            if isinstance(register_response, AM.RegisterEngineReplyMsg) and not register_response.secret_match:
                logger.error("Secret supplied by engine does not match aggregator secret")
            return None
        logger.debug(f"Aggregator accepted registration and issued engine_id: {register_response.engine_id}")
        return register_response.engine_id

    def assign_sequence_number(self, message: EM.EngineMessage | EM.RegisterEngineMsg):
        # only assign sequence number once (otherwise buffered messages would be reassigned)
        if message.sequence_number == -1:
            self._sequence_number += 1
            message.sequence_number = self._sequence_number



import logging
import math
from typing import List

from antlr4 import ParserRuleContext

from openpectus.lang.grammar.codegen.pcodeParser import pcodeParser
from openpectus.lang.grammar.codegen.pcodeListener import pcodeListener

from openpectus.lang.model.pprogram import (
    PCommandWithDuration,
    PDuration,
    PInstError,
    PNode,
    PProgram,
    PInstruction,
    PBlock,
    PEndBlock,
    PEndBlocks,
    PWatch,
    PAlarm,
    PMark,
    PCommand,
    PError,
    PBlank,
    PCondition,
    PComment,
    PErrorInstruction
)

INDENTATION_SPACES = 4

logger = logging.getLogger(__name__)


class PProgramBuilder(pcodeListener):
    def __init__(self) -> None:
        super().__init__()

        self.program: PProgram = PProgram()
        self.stack: List[PNode] = []
        """ Stack of scopes. """

        self.instruction: PInstruction | None = None
        """ The current instruction being built. """

        self.prev_instruction: PInstruction | None = None
        """ The previous instruction. """

        self.instruction_start: InstructionStart | None = None
        """ Start position of the current instruction. """

        self.instruction_error: PError | None = None
        """ Error to be attached to the current instruction. """

        self.expect_indent: bool = False
        """ Flag raised by instructions that expect the next instruction to be additionally indented """

    @property
    def scope(self) -> PNode:
        """ Returns the current program or instruction scope. """
        if not self.has_scope():
            raise LookupError("No scope has been set")
        return self.stack[-1]

    def has_scope(self) -> bool:
        return len(self.stack) > 0

    def push_scope(self, new_scope: PInstruction, ctx: ParserRuleContext) -> PInstruction:
        """ Add instruction as the new inner scope.
        """
        self.stack.append(new_scope)
        self.expect_indent = True

        return new_scope

    def pop_scope(self) -> PNode:
        """ Leave the current scope and return to outer scope.
        """
        if len(self.stack) < 1:
            raise ValueError("Stack underflow")
        parent_scope = self.stack.pop()
        return parent_scope

    def get_program(self) -> PProgram:
        return self.program


    def enterProgram(self, ctx: pcodeParser.ProgramContext):
        # set program as scope
        self.program.line = ctx.start.line
        self.program.indent = ctx.start.column
        self.stack.append(self.program)

    def enterInstruction(self, ctx: pcodeParser.InstructionContext):
        instruction_text = ctx.getText()
        instruction_text = "" if instruction_text is None else instruction_text.strip()

        def is_blank_or_comment():
            if instruction_text == "":
                return True
            elif instruction_text.startswith("#"):
                return True
            return False

        if not isinstance(self.instruction, (PBlank, PComment)):
            self.prev_instruction = self.instruction
        self.instruction = None
        self.instruction_error = None
        self.instruction_start = InstructionStart(ctx.start.line, ctx.start.column)

        indent = int(ctx.start.column)
        if math.remainder(indent, INDENTATION_SPACES) != 0:
            # validate indentation as multiple of INDENTATION_SPACES
            self.instruction_error = PError(f'Bad indentation. Should be a multiple of {INDENTATION_SPACES}')
        else:
            # validate scope indentation
            assert isinstance(self.scope.indent, int), "Expect indent always set"
            prev_indent = 0 \
                if self.prev_instruction is None or self.prev_instruction.indent is None \
                else self.prev_instruction.indent
            # if outdented, just pop to relevant parent scope
            # unless it's a blank or comment
            if not is_blank_or_comment():
                if indent < prev_indent:
                    x = indent
                    while (x < prev_indent):
                        self.pop_scope()
                        x += INDENTATION_SPACES
                # if additional indent, check that this matches expectation set by previous instruction
                elif indent > prev_indent:
                    if not self.expect_indent:
                        self.instruction_error = PError(
                            f"Bad indentation. Additional indentation unexpected. Expected {prev_indent} " +
                            " spaces of indentation")
                        # clear the invalid extra indentation to avoid errors in following instructions
                        self.instruction_start.indent = prev_indent
                    elif indent != prev_indent + INDENTATION_SPACES:
                        self.instruction_error = PError(
                            f"Bad indentation. Expected {prev_indent + INDENTATION_SPACES} spaces of indentation")
                    self.expect_indent = False
                # if no change in indentation but one was expected, set error
                elif self.expect_indent:
                    self.instruction_error = PError(
                        f"Bad indentation. Expected additional indentation to {prev_indent + INDENTATION_SPACES} spaces")

    def exitInstruction(self, ctx: pcodeParser.InstructionContext):
        # attach any error to current instruction
        if self.instruction_error is not None:
            if self.instruction is None:
                self.instruction = PErrorInstruction(self.scope, ctx.getText())
            elif isinstance(self.instruction, PBlank):
                # skip indentation errors in blank lines
                # TODO use error code instead
                if 'indentation' not in (self.instruction_error.message or ""):
                    self.instruction.add_error(self.instruction_error)
            else:
                self.instruction.add_error(self.instruction_error)

        # attach start position
        assert self.instruction_start is not None
        if self.instruction is None:
            self.instruction = PErrorInstruction(self.scope, ctx.getText())

        self.instruction.line = self.instruction_start.line
        self.instruction.indent = self.instruction_start.indent

    def enterBlock(self, ctx: pcodeParser.BlockContext):
        self.instruction = PBlock(self.scope)
        self.push_scope(self.instruction, ctx)

    def enterBlock_name(self, ctx: pcodeParser.Block_nameContext):
        if isinstance(self.scope, PBlock) and isinstance(self.instruction, PBlock):
            self.scope.name = ctx.getText()

    def enterEnd_block(self, ctx: pcodeParser.End_blockContext):
        self.instruction = PEndBlock(self.scope)

    def enterEnd_blocks(self, ctx: pcodeParser.End_blocksContext):
        self.instruction = PEndBlocks(self.scope)

    def enterCommand(self, ctx: pcodeParser.CommandContext):
        self.instruction = PCommand(self.scope)

    def enterCommand_name(self, ctx: pcodeParser.Command_nameContext):
        assert isinstance(self.instruction, PCommand)
        self.instruction.name = ctx.getText()

    def enterCommand_args(self, ctx: pcodeParser.Command_argsContext):
        assert isinstance(self.instruction, PCommand)
        self.instruction.args = ctx.getText()

    def enterWatch(self, ctx: pcodeParser.WatchContext):
        self.instruction = PWatch(self.scope)
        self.push_scope(self.instruction, ctx)

    def enterCondition(self, ctx: pcodeParser.ConditionContext):
        assert isinstance(self.scope, PWatch | PAlarm)
        self.scope.condition = PCondition(ctx.getText())

    def enterCompare_op(self, ctx: pcodeParser.Compare_opContext):
        if isinstance(self.scope, (PAlarm, PWatch)) and self.scope.condition:
            self.scope.condition.op = ctx.getText()

    def enterCondition_lhs(self, ctx: pcodeParser.Condition_lhsContext):
        if isinstance(self.scope, (PAlarm, PWatch)) and self.scope.condition:
            self.scope.condition.lhs = ctx.getText()

    def enterCondition_rhs(self, ctx: pcodeParser.Condition_rhsContext):
        if isinstance(self.scope, (PAlarm, PWatch)) and self.scope.condition:
            self.scope.condition.rhs = ctx.getText()

    def enterAlarm(self, ctx: pcodeParser.AlarmContext):
        self.instruction = PAlarm(self.scope)
        self.push_scope(self.instruction, ctx)

    def enterIncrement_rc(self, ctx: pcodeParser.Increment_rcContext):
        if self.instruction is None:
            self.instruction = PCommand(self.scope)
            self.instruction.name = ctx.getText()

    def enterRestart(self, ctx: pcodeParser.RestartContext):
        if self.instruction is None:
            self.instruction = PCommand(self.scope)
            self.instruction.name = ctx.getText()

    def enterStop(self, ctx: pcodeParser.StopContext):
        if self.instruction is None:
            self.instruction = PCommand(self.scope)
            self.instruction.name = ctx.getText()

    def enterPause(self, ctx: pcodeParser.PauseContext):
        assert self.instruction is None
        self.instruction = PCommandWithDuration(self.scope)
        self.instruction.name = ctx.PAUSE().getText()

    def enterHold(self, ctx: pcodeParser.HoldContext):
        assert self.instruction is None
        self.instruction = PCommandWithDuration(self.scope)
        self.instruction.name = ctx.HOLD().getText()

    def enterWait(self, ctx: pcodeParser.WaitContext):
        assert self.instruction is None
        self.instruction = PCommandWithDuration(self.scope)
        self.instruction.name = ctx.WAIT().getText()

    def enterDuration(self, ctx: pcodeParser.DurationContext):
        assert isinstance(self.instruction, PCommandWithDuration)
        self.instruction.duration = PDuration(ctx.getText())

    def enterMark(self, ctx: pcodeParser.MarkContext):
        self.instruction = PMark(self.scope)

    def enterMark_name(self, ctx: pcodeParser.Mark_nameContext):
        assert isinstance(self.instruction, PMark)
        self.instruction.name = ctx.getText()

    def enterTimeexp(self, ctx: pcodeParser.TimeexpContext):
        # NOTE: there is currently a mismatch between PInstruction which have 'time' and the grammar
        # where not all instructions have 'time'.
        assert isinstance(self.instruction, PInstruction)
        time = ctx.getText()
        self.instruction.time = None if time is None or time.strip() == '' else float(time)

    def enterComment(self, ctx: pcodeParser.CommentContext):
        if self.instruction is None:
            self.instruction = PComment(self.scope)

        if self.instruction.comment == "":
            comment = ctx.getText()
            comment = "" if comment is None else comment[1:]
            self.instruction.comment = comment
        else:
            # account for multiple comments on same line
            self.instruction.comment += ctx.getText()

    def enterBlank(self, ctx: pcodeParser.BlankContext):
        self.instruction = PBlank(self.scope)

    def enterInst_error(self, ctx: pcodeParser.Inst_errorContext):
        if self.instruction is not None:
            self.instruction._inst_error = PInstError(ctx.getText())
        else:
            logger.warning("In enterInst_error, instruction was none. This should not happen")

    def enterError(self, ctx: pcodeParser.ErrorContext):
        pass

class InstructionStart():
    def __init__(self, line: int, indent: int) -> None:
        self.line = line
        self.indent = indent



from __future__ import annotations
import logging
import math
from datetime import datetime
from enum import StrEnum, auto

import openpectus.protocol.models as Mdl
from pydantic import BaseModel

logger = logging.getLogger(__name__)

# aliases so users of this file don't have to know about the protocol models
RunLogLine = Mdl.RunLogLine
RunLog = Mdl.RunLog
ControlState = Mdl.ControlState
Method = Mdl.Method
MethodLine = Mdl.MethodLine
MethodState = Mdl.MethodState
ReadingCommand = Mdl.ReadingCommand
ReadingInfo = Mdl.ReadingInfo
CommandInfo = Mdl.CommandInfo
TagValue = Mdl.TagValue
PlotColorRegion = Mdl.PlotColorRegion
PlotAxis = Mdl.PlotAxis
SubPlot = Mdl.SubPlot
PlotConfiguration = Mdl.PlotConfiguration
ErrorLogEntry = Mdl.ErrorLogEntry
ErrorLog = Mdl.ErrorLog
SystemStateEnum = Mdl.SystemStateEnum
TagDirection = Mdl.TagDirection


class AggregatedErrorLogEntry(BaseModel):
    message: str
    created_time: float
    severity: int
    occurrences: int = 1

    @staticmethod
    def from_entry(entry: ErrorLogEntry):
        return AggregatedErrorLogEntry(
            message=entry.message,
            created_time=entry.created_time,
            severity=entry.severity,
            occurrences=1
        )


class AggregatedErrorLog(BaseModel):
    entries: list[AggregatedErrorLogEntry]

    @staticmethod
    def empty() -> AggregatedErrorLog:
        return AggregatedErrorLog(entries=[])

    def clear(self):
        self.entries.clear()

    def aggregate_with(self, error_log: ErrorLog):
        """ Append the given log entries to the current log entries while deduplicating and aggregating entries. """
        latest = self.entries[-1] if len(self.entries) > 0 else None
        for entry in error_log.entries:
            assert entry is not None, "log entry should not be None"
            if latest is not None and entry.message == latest.message and entry.severity == latest.severity:
                if latest.created_time < entry.created_time:
                    latest.created_time = entry.created_time
                    latest.occurrences += 1
                elif latest.created_time == entry.created_time:
                    logger.warning(f"Duplicate log entry with same created_time: {entry.message}")
                else:
                    logger.warning(f"Duplicate log entry with earlier created_time: {entry.message}")
            else:
                latest = AggregatedErrorLogEntry.from_entry(entry)
                self.entries.append(latest)


class ChannelStatusEnum(StrEnum):
    Unknown = auto()
    Connected = auto()
    Registered = auto()
    Disconnected = auto()


TagValueType = int | float | str | None
""" Represents the possible types of a tag value"""


class TagsInfo(BaseModel):
    map: dict[str, Mdl.TagValue]

    def get(self, tag_name: str):
        return self.map.get(tag_name)

    def upsert(self, tag_value: Mdl.TagValue) -> bool:
        current = self.map.get(tag_value.name)
        if current is None:
            self.map[tag_value.name] = tag_value
            return True  # was inserted
        else:
            if current.value_unit != tag_value.value_unit:
                logger.warning(f"Tag '{tag_value.name}' changed unit from '{current.value_unit}' to " +
                               f"'{tag_value.value_unit}'. This is unexpected")
            if __debug__:
                if current.tick_time > tag_value.tick_time:
                    cur_time = datetime.fromtimestamp(current.tick_time).strftime("%H:%M:%S")
                    new_time = datetime.fromtimestamp(tag_value.tick_time).strftime("%H:%M:%S")
                    logger.warning(f"Tag '{tag_value.name}' was updated with an earlier time " +
                                   f"{new_time} than the current time {cur_time}.")
            current.value = tag_value.value
            current.value_formatted = tag_value.value_formatted
            current.tick_time = tag_value.tick_time
            return False  # was updated

    def __str__(self) -> str:
        return f"TagsInfo({','.join(self.map.keys())})"

    def get_last_modified_time(self) -> datetime | None:
        if len(self.map.keys()) == 0:
            return None
        max_tick_time = max(t.tick_time for t in self.map.values())
        return datetime.fromtimestamp(max_tick_time)


class RunData(BaseModel):
    """ Represents data that strictly belongs in a specific run. """

    run_id: str
    run_started: datetime
    runlog: RunLog = RunLog.empty()
    latest_persisted_tick_time: float | None = None
    """ The time assigned to the last persisted set of values """
    latest_tag_time: float | None = None
    """ The time of the latest tag update, persisted or not """
    interrupted_by_error: bool = False

    def __str__(self) -> str:
        return f"RunData({self.run_id=}, {self.run_started=}, " + \
               f"interrupted_by_error:{self.interrupted_by_error})"

    @staticmethod
    def empty(run_id: str, run_started: datetime) -> RunData:
        return RunData(run_id=run_id, run_started=run_started)


class EngineData():
    """ Data stored by aggregator for each connected engine. """

    # Note: Not a BaseModel subclass since it doesn't need to be and BaseModel apparently
    # has a bug concerning the way we use _run_data, run_data and has_run().

    def __init__(self, engine_id: str, computer_name: str, engine_version: str, uod_name: str, uod_author_name: str,
                 uod_author_email: str, uod_filename: str, location: str,
                 hardware_str: str = "N/A", data_log_interval_seconds: float = math.inf) -> None:
        self.engine_id: str = engine_id
        self.computer_name: str = computer_name
        self.engine_version: str = engine_version
        self.uod_name: str = uod_name
        self.uod_author_name: str = uod_author_name
        self.uod_author_email: str = uod_author_email
        self.uod_filename: str = uod_filename
        self.location: str = location
        self.readings: list[Mdl.ReadingInfo] = []
        """ Contains the definition of the engine's process values. """
        self.commands: list[Mdl.CommandInfo] = []
        """ Contains the uod commands that are not related to a process value. """
        self.tags_info: TagsInfo = TagsInfo(map={})
        """ Contains the most current tag values. """
        self.control_state: ControlState = ControlState(is_running=False, is_holding=False, is_paused=False)
        self.method: Method = Method.empty()
        self._run_data: RunData | None = None
        self.error_log: AggregatedErrorLog = AggregatedErrorLog.empty()
        self.method_state: MethodState = MethodState.empty()
        self.plot_configuration: PlotConfiguration = PlotConfiguration.empty()
        self.contributors: set[str] = set()
        self.required_roles: set[str] = set()
        self.hardware_str: str = hardware_str
        self.data_log_interval_seconds: float = data_log_interval_seconds

    @property
    def run_data(self) -> RunData:
        if self._run_data is None:
            raise ValueError("Run data is not set. Guard access with call to has_run()")
        return self._run_data

    @run_data.setter
    def run_data(self, value: RunData):
        self._run_data = value

    @property
    def runtime(self):
        return self.tags_info.get(Mdl.SystemTagName.RUN_TIME)

    @property
    def system_state(self):
        return self.tags_info.get(Mdl.SystemTagName.SYSTEM_STATE)

    def has_run(self) -> bool:
        """ Determine whether run_data is available, that is, whether a run is active. """
        return self._run_data is not None

    def reset_run(self):
        """ Clear all data associated with a run, control_state, error_log, method state (not method content) and run_data,
        including run log and run_id. Call when run is complete to get ready for a new run. """
        self._run_data = None
        self.control_state = ControlState(is_running=False, is_holding=False, is_paused=False)
        self.error_log.clear()
        self.method_state = MethodState.empty()
        self.contributors = set()

    def __str__(self) -> str:
        return f"EngineData(engine_id:{self.engine_id}, control_state':{self.control_state})"



import logging
from typing import Any, Iterable, Dict

from openpectus.engine.hardware import (
    HardwareLayerBase,
    HardwareLayerException,
    Register,
)

logger = logging.getLogger(__name__)


class Composite_Hardware(HardwareLayerBase):
    """ Combine multiple hardware layers into one by
        assigning the hardware layer on a per-register basis.

        Example pseudo-code:
        hwl_A = Hardware_1()
        hwl_B = Hardware_2()
        registers = dict()
        registers["Reg 1"] = Register("Reg 1", RegisterDirection.Both, hardware=hwl_A)
        registers["Reg 2"] = Register("Reg 2", RegisterDirection.Both, hardware=hwl_B)
        hwl = Composite_Hardware()
        hwl.registers = registers
        with hwl:
            hwl.read(registers["Reg 1"])"""
    def _underlying_hardwares(self) -> list[HardwareLayerBase]:
        """ Construct list of unique hardwares from list of registers """
        return list(dict.fromkeys(r.options["hardware"] for r in self.registers.values()))

    def validate_offline(self):
        for r in self.registers.values():
            if "hardware" not in r.options:
                raise HardwareLayerException(f"Hardware layer undefined for register {r}.")
            if not issubclass(type(r.options["hardware"]), HardwareLayerBase):
                raise HardwareLayerException((f"Hardware layer {r.options['hardware']} "
                                              f"specified for register {r} is not a valid "
                                              f"hardware layer."))
            if isinstance(r.options["hardware"], type(self)):
                raise HardwareLayerException((f"Hardware layer {r.options['hardware']} "
                                              f"specified for register {r} is itself a "
                                              f"composite hardware layer. This is not allowed."))
        for hardware in self._underlying_hardwares():
            hardware.validate_online()

    def validate_online(self):
        for hardware in self._underlying_hardwares():
            hardware.validate_online()

    def read(self, r: Register) -> Any:
        return r.options["hardware"].read(r)

    def read_batch(self, registers: Iterable[Register]) -> list[Any]:
        """ Divide list of registers by the hardware they belong to,
        perform the batch read, gather the results and order them
        like the list of registers. """
        registers_by_hardware: Dict[HardwareLayerBase, list[Register]] = dict()
        read_by_register: Dict[Register, Any] = dict()
        for r in registers:
            if r.options["hardware"] not in registers_by_hardware:
                registers_by_hardware[r.options["hardware"]] = [r]
            else:
                registers_by_hardware[r.options["hardware"]].append(r)
        for hardware, registers_belonging_to_hardware in registers_by_hardware.items():
            reads = hardware.read_batch(registers_belonging_to_hardware)
            for register, read in zip(registers_belonging_to_hardware, reads):
                read_by_register[register] = read
        return [read_by_register[register] for register in registers]

    def write(self, value: Any, r: Register):
        r.options["hardware"].write(value, r)

    def write_batch(self, values: Iterable[Any], registers: Iterable[Register]):
        """ Divide list of registers by the hardware they belong to,
        perform the batch write with values in same order as the registers. """
        registers_by_hardware: Dict[HardwareLayerBase, list[Register]] = dict()
        write_value_by_register: Dict[Register, Any] = dict()
        for v, r in zip(values, registers):
            if r.options["hardware"] not in registers_by_hardware:
                registers_by_hardware[r.options["hardware"]] = [r]
            else:
                registers_by_hardware[r.options["hardware"]].append(r)
            write_value_by_register[r] = v
        for hardware, registers_belonging_to_hardware in registers_by_hardware.items():
            values_to_hardware = [write_value_by_register[r] for r in registers_belonging_to_hardware]
            hardware.write_batch(values_to_hardware, registers_belonging_to_hardware)

    def connect(self):
        for hardware in self._underlying_hardwares():
            hardware.connect()
        super().connect()

    def disconnect(self):
        for hardware in self._underlying_hardwares():
            hardware.disconnect()
        super().disconnect()

    def __str__(self):
        return f"Composite_Hardware(hardwares={self._underlying_hardwares()})"



import logging
import functools
import itertools
from typing import Any, Sequence
import asyncua
import asyncua.ua
import asyncua.sync
import asyncua.client.ua_client

from openpectus.engine.hardware import (
    HardwareLayerBase,
    Register,
    RegisterDirection,
    HardwareLayerException,
)

logger = logging.getLogger(__name__)

OPCUA_Types = asyncua.ua.VariantType


def _opcua_parent_path(path):
    """ Returns the parent path for an OPC-UA path.

    > path = "Objects/2:FT01/2:Totalizer/2:Total 1"
    > self._parent_path(path)
    "Objects/2:FT01/2:Totalizer"
    """
    if '/' in path:
        return path.rsplit('/', 1)[0]
    else:
        return path


def batched(iterable: Sequence[Any], n: int):
    """ Batch data into tuples of length n. The last batch may be shorter.
    Sourced from https://docs.python.org/3.11/library/itertools.html#itertools.batched.
    In Python 3.12 this will become a part of the itertools package
    and this function can be replaced by 'from itertools import batched'."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


#  Monkeypatch asyncua library
# The asyncua OPC-UA client launches a thread class called
# ThreadLoop which is unfortunately not configured to be daemonic.
# This can cause a python instance to hang if the client
# connection is not deliberately disconnected.
# A PR to asyncua petitioning them to set it daemonic in the
# first place has been submitted, see
# https://github.com/FreeOpcUa/opcua-asyncio/pull/1555.
asyncua.sync.ThreadLoop.daemon = True


class OPCUA_Hardware(HardwareLayerBase):
    """ Represents OPCUA hardware layer. """
    def __init__(self, host: str) -> None:
        super().__init__()
        self.host: str = host
        self._client: asyncua.sync.Client = asyncua.sync.Client(self.host)
        # Siemens default maximum values
        self._max_nodes_per_read: int = 1000
        self._max_nodes_per_write: int = 1000

    def _browse_opcua_name_space_depth_first_until_path_is_valid(
            self, path: str) -> tuple[str, list[asyncua.ua.uatypes.QualifiedName]]:
        """ Given a path to a node that does not exist, find the closest
        parent that does.

        > path_to_nonexistant_node = "Objects/2:FT01/2:Totalizer/2:Totaal 1"
        > # User made a spelling mistake at the final node above. Oh dear...
        > # Suppose that "Objects/2:FT01/2:Totalizer" does exist though
        > self._browse_opcua_name_space_depth_first_until_path_is_valid(
        > path_to_nonexistant_node)
        [QualifiedName(NamespaceIndex=2, Name='Total 1'),
        QualifiedName(NamespaceIndex=2, Name='Total 2'),
        QualifiedName(NamespaceIndex=2, Name='Total 3')]
        """
        while True:
            path = _opcua_parent_path(path)
            if '/' in path:  # We are still not at the "Objects" node
                try:
                    children = self._client.nodes.root.get_child(path).get_children()
                    break
                except asyncua.ua.uaerrors._auto.BadNoMatch:
                    pass
            else:  # We've reached the "Objects" node
                children = self._client.get_objects_node().get_children()
                break
        children = [child.read_browse_name() for child in children]
        return (path, children,)

    def _query_server_operation_limits(self):
        """ Ask the server about it's operational capabilities. """
        registers = [
            Register("max_nodes_per_read",
                     RegisterDirection.Read,
                     path="Objects/0:Server/0:ServerCapabilities/0:OperationLimits/0:MaxNodesPerRead"),
            Register("max_nodes_per_write",
                     RegisterDirection.Read,
                     path="Objects/0:Server/0:ServerCapabilities/0:OperationLimits/0:MaxNodesPerWrite"),
        ]
        self._max_nodes_per_read, self._max_nodes_per_write = self.read_batch(registers)

    def validate_offline(self):
        for r in self.registers.values():
            # Check that path is specified
            if "path" not in r.options:
                raise HardwareLayerException((f"OPC-UA hardware layer requires specification "
                                              f"of a path to the register value. Please "
                                              f"specify path for register {r}."))
            # Check that path is sensible
            if "/" not in r.options["path"]:
                if not r.options["path"] == "Objects":
                    raise HardwareLayerException(f"Invalid OPC-UA node path '{r.options['path']}' for register {r}.")
            else:
                for browse_name in r.options["path"].split("/")[1:]:
                    if not browse_name.count(":") == 1:
                        raise HardwareLayerException((f"Invalid OPC-UA node path '{r.options['path']}' "
                                                      f"for register {r}. Node browse name "
                                                      f"'{browse_name}' specifically is not valid."))
            # If a type is specified, check that it is sensible
            if "type" in r.options and not isinstance(r.options["type"], asyncua.ua.VariantType):
                raise HardwareLayerException((f"Invalid type '{r.options['type']}' "
                                              f"specified for register {r}. Type must "
                                              f"be an asyncua.ua.VariantType."))

    def validate_online(self):
        for r in self.registers.values():
            # Get the node
            node = self._register_to_node(r)
            # Read the type of the node
            opcua_type = node.read_data_type_as_variant_type()
            # Verify that register type and actual OPC-UA type match
            if 'type' in r.options:
                register_type = r.options["type"]
                if register_type is not opcua_type:
                    raise HardwareLayerException((f"Disparity between type '{OPCUA_Types(register_type).name}' "
                                                  f"for register {r} and actual OPC-UA type "
                                                  f"'{OPCUA_Types(opcua_type).name}'."))
            # Verify access level
            access_level = node.get_access_level()
            if (
               RegisterDirection.Read in r.direction and
               asyncua.ua.AccessLevel.CurrentRead not in access_level
               ) or (
               RegisterDirection.Write in r.direction and
               asyncua.ua.AccessLevel.CurrentWrite not in access_level
               ):
                raise HardwareLayerException((f"Disparity between register direction '{r.direction}' "
                                              f"for register {r} and OPC-UA access level "
                                              f"'{access_level}."))

    @functools.cache
    def _register_to_node(self, r: Register) -> asyncua.sync.SyncNode:
        """ Resolving a NodeId from a path requires an OPC-UA call.
        The NodeId's are fixed so it makes sense to resolve them once
        and cache the result for future use. """
        path = r.options["path"]
        try:
            node_id = self._client.nodes.root.get_child(path)
        except asyncua.ua.uaerrors._auto.BadNoMatch:
            # If the node cannot be found then we try to help the
            # person writing the UOD debug by browsing the OPC-UA
            # namespace for them.
            valid_path, children = self._browse_opcua_name_space_depth_first_until_path_is_valid(path)
            children_string = ", ".join([f"{child.NamespaceIndex}:{child.Name}" for child in children])
            raise HardwareLayerException((f"Node at register {r} path {path} "
                                          f"cannot be found. The closest parent "
                                          f"node is '{valid_path}' with children: {children_string}"))
        except asyncua.sync.ThreadLoopNotRunning:
            raise HardwareLayerException("OPC-UA client is closed.")
        return node_id

    def _registers_to_nodes(self, registers: Sequence[Register]) -> list[asyncua.sync.SyncNode]:
        node_ids = []
        for r in registers:
            node_ids.append(self._register_to_node(r))
        return node_ids

    def read(self, r: Register) -> Any:
        if RegisterDirection.Read not in r.direction:
            raise HardwareLayerException(f"Attempt to read unreadable register {r}.")
        try:
            opcua_node = self._register_to_node(r)
            return opcua_node.read_value()
        except ConnectionError:
            raise HardwareLayerException(f"Not connected to {self.host}")
        except asyncua.sync.ThreadLoopNotRunning:
            logger.error("Thread loop not running", exc_info=True)
            raise HardwareLayerException("OPC-UA client is closed.")

    def read_batch(self, registers: Sequence[Register]) -> list[Any]:
        """ Read batch of register values with a single OPC-UA call. """
        for r in registers:
            if RegisterDirection.Read not in r.direction:
                raise HardwareLayerException(f"Attempt to read unreadable register {r}.")
        values = []
        try:
            for register_batch in batched(registers, self._max_nodes_per_read):
                nodes = self._registers_to_nodes(register_batch)
                values += self._client.read_values(nodes)
        except ConnectionError:
            raise HardwareLayerException(f"Not connected to {self.host}")
        except asyncua.sync.ThreadLoopNotRunning:
            logger.error("Thread loop not running", exc_info=True)
            raise HardwareLayerException("OPC-UA client is closed.")
        return values

    def _coerce_value_to_proper_type(self, value: Any, r: Register) -> Any:
        if "type" not in r.options:
            return value

        coercion_mapping = {
            OPCUA_Types.Boolean: lambda x: bool(x),
            OPCUA_Types.Byte: lambda x: int(x),
            OPCUA_Types.Int16: lambda x: int(x),
            OPCUA_Types.UInt16: lambda x: 0 if x < 0 else int(x),
            OPCUA_Types.Int32: lambda x: int(x),
            OPCUA_Types.UInt32: lambda x: 0 if x < 0 else int(x),
            OPCUA_Types.Int64: lambda x: int(x),
            OPCUA_Types.UInt64: lambda x: 0 if x < 0 else int(x),
            OPCUA_Types.Float: lambda x: float(x),
            OPCUA_Types.Double: lambda x: float(x),
        }

        limits_check_mapping = {
            OPCUA_Types.Boolean: lambda x: x in [True, False],
            OPCUA_Types.Byte: lambda x: x >= 0 and x < 256,
            OPCUA_Types.Int16: lambda x: x >= -(2**15-1) and x < (2**15-1),
            OPCUA_Types.UInt16: lambda x: x >= 0 and x < (2**16-1),
            OPCUA_Types.Int32: lambda x: x >= -(2**31-1) and x < (2**31-1),
            OPCUA_Types.UInt32: lambda x: x >= 0 and x < (2**32-1),
            OPCUA_Types.Int64: lambda x: x >= -(2**63-1) and x < (2**63-1),
            OPCUA_Types.UInt64: lambda x: x >= 0 and x < (2**64-1),
        }

        coercion_function = coercion_mapping.get(r.options["type"], None)
        limits_check_function = limits_check_mapping.get(r.options["type"], None)

        coerced_value = value
        if coercion_function:
            try:
                coerced_value = coercion_function(value)
            except Exception:
                raise HardwareLayerException((f"Value '{value}' could not be coerced to type '{r.options['type']}' "
                                              f"when writing register {r}."))
            if coerced_value is not value:
                logger.warning((f"Value '{value}' was coerced to '{coerced_value}' "
                                f"when writing register {r}."))

        if limits_check_function:
            if not limits_check_function(coerced_value):
                raise HardwareLayerException((f"Value '{coerced_value}' does not satisfy the limits that its "
                                              f"type '{r.options['type']}' imposes when writing register {r}."))

        return coerced_value

    def write(self, value: Any, r: Register):
        if RegisterDirection.Write not in r.direction:
            raise HardwareLayerException(f"Attempt to write unwritable register {r}.")
        try:
            opcua_node = self._register_to_node(r)
            value = self._coerce_value_to_proper_type(value, r)
            if 'type' in r.options:
                value = asyncua.ua.Variant(value, r.options['type'])
            data_value = asyncua.ua.DataValue(value)
            opcua_node.write_attribute(asyncua.ua.AttributeIds.Value, data_value)
        except ConnectionError:
            raise HardwareLayerException(f"Not connected to {self.host}")
        except asyncua.sync.ThreadLoopNotRunning:
            logger.error("Thread loop not running", exc_info=True)
            raise HardwareLayerException("OPC-UA client is closed.")

    def write_batch(self, values: Sequence[Any], registers: Sequence[Register]):
        """ Write batch of register values with a single OPC-UA call. """
        for r in registers:
            if RegisterDirection.Write not in r.direction:
                raise HardwareLayerException(f"Attempt to write unwritable register {r}.")
        try:
            node_ids = []
            nodes = self._registers_to_nodes(registers)
            node_ids = [node.nodeid for node in nodes]
            data_values = []
            for r, value in zip(registers, values):
                value = self._coerce_value_to_proper_type(value, r)
                if 'type' in r.options:
                    value = asyncua.ua.Variant(value, r.options['type'])
                data_values.append(asyncua.ua.DataValue(value))
            # This is quite convoluted because asyncua.sync does not offer a convenience method.
            # The approach taken to get a synchronous "write_attributes" method is documented
            # in opcua-asyncua asyncua/sync.py source code:
            # https://github.com/FreeOpcUa/opcua-asyncio/blob/master/asyncua/sync.py
            write_attributes = asyncua.sync.sync_uaclient_method(
                               asyncua.client.ua_client.UaClient.write_attributes)(self._client)
            for node_id_batch, data_value_batch in zip(batched(node_ids, self._max_nodes_per_write),
                                                       batched(data_values, self._max_nodes_per_write)):
                write_attributes(node_id_batch, data_value_batch, asyncua.ua.AttributeIds.Value)
        except ConnectionError:
            raise HardwareLayerException(f"Not connected to {self.host}")
        except asyncua.sync.ThreadLoopNotRunning:
            logger.error("Thread loop not running", exc_info=True)
            raise HardwareLayerException("OPC-UA client is closed.")

    def connect(self):
        """ Connect to OPC-UA server. """
        logger.info(f"Attempting to connect to {self.host}")
        if self._client:
            # catching these exceptions make reconnect able to just try
            # again which seems to work
            try:
                self._client.disconnect()
            except asyncua.sync.ThreadLoopNotRunning:
                pass
            except RuntimeError:
                pass
            # Remove potential stale references to nodes on re-connect
            self._register_to_node.cache_clear()
            del self._client
        self._client = asyncua.sync.Client(self.host)
        try:
            self._client.connect()
        except ConnectionRefusedError:
            logger.error(f"Unable to connect to {self.host}")
            raise HardwareLayerException(f"Unable to connect to {self.host}")
        logger.info(f"Connected to {self.host}")
        self._query_server_operation_limits()
        super().connect()

    def disconnect(self):
        """ Disconnect hardware. """
        logger.info(f"Disconnecting from {self.host}")
        if self._client:
            self._client.disconnect()
        super().disconnect()

    def __str__(self):
        return f"OPCUA_Hardware(host={self.host})"



"""removed ProcessUnits table

Revision ID: bcda1e0dd461
Revises: d53dfbe96f00
Create Date: 2024-11-27 11:22:00.975364

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'bcda1e0dd461'
down_revision: Union[str, None] = 'd53dfbe96f00'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('ProcessUnits')
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('ProcessUnits',
    sa.Column('engine_id', sa.VARCHAR(), nullable=False),
    sa.Column('state', sa.VARCHAR(), nullable=False),
    sa.Column('location', sa.VARCHAR(), nullable=False),
    sa.Column('runtime_msec', sa.INTEGER(), nullable=False),
    sa.Column('current_user_role', sa.VARCHAR(), nullable=False),
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###



import asyncio
import json
import logging
from typing import Callable, Awaitable, Any

from fastapi import APIRouter, FastAPI, Request
from fastapi_websocket_rpc import RpcMethodsBase, RpcChannel, WebsocketRPCEndpoint
from fastapi_websocket_rpc.schemas import RpcResponse

import openpectus.protocol.aggregator_messages as AM
import openpectus.protocol.engine_messages as EM
import openpectus.protocol.messages as M
from openpectus.protocol.dispatch_interface import AGGREGATOR_REST_PATH, AGGREGATOR_RPC_WS_PATH, EngineMessageType
from openpectus.protocol.exceptions import ProtocolException
from openpectus.protocol.serialization import deserialize, serialize


logger = logging.getLogger(__name__)


RegisterHandler = Callable[[EM.RegisterEngineMsg], Awaitable[AM.RegisterEngineReplyMsg]]
""" Specific handler for register messages from engine """

EngineConnectHandler = Callable[[str], Awaitable[None]]
""" Specific handler for engine disconnect notifications from dispatcher """

EngineDisconnectHandler = Callable[[str], Awaitable[None]]
""" Specific handler for engine disconnect notifications from dispatcher """

AggregatorMessageHandler = Callable[[Any], Awaitable[M.MessageBase]]
""" Handler in aggregator that handles engine messages of a given type """

WEBSOCKET_RPC_TIMEOUT_SECS: float | None = 2.0


class AggregatorDispatcher():
    """
    Aggregator dispatcher for the Aggregator-Engine Protocol using REST + WebSocket RPC.
    Allows receiving message via HTTP POST and sending messages via JSON-RPC.
    """

    def __init__(self) -> None:
        self._handlers: dict[type, AggregatorMessageHandler] = {}
        self._register_handler: RegisterHandler | None = None
        self._connect_handler: EngineConnectHandler | None = None
        self._disconnect_handler: EngineDisconnectHandler | None = None
        self.router = APIRouter(tags=["aggregator"])
        self._engine_id_channel_map: dict[str, RpcChannel] = {}
        self._on_client_connect_tasks = set()
        # WebsockeRPCEndpoint has wrong types for its on_connect and on_disconnect.
        # It should be List[Callable[[RpcChannel], Awaitable[Any]]] instead of List[Coroutine]
        # See https://github.com/permitio/fastapi_websocket_rpc/issues/30

        class AggregatorRpcMethods(RpcMethodsBase):
            def __init__(self, dispatcher: AggregatorDispatcher):
                super().__init__()
                self.disp = dispatcher

            # Why does this work with return type?? In engine_dispatcher
            # it fails if we do this. At least we know how to fix it, should it should fail.
            async def dispatch_message_async(self, message_json: dict[str, Any]) -> str:
                """ Handle RCP call from engine. Dispatch message to registered handler. """
                try:
                    message = deserialize(message_json)
                    assert isinstance(message, EM.EngineMessage)  # only allow EngineMessage, Register is via rest
                except Exception:
                    logger.error("Dispatch failed. Message deserialization failed.", exc_info=True)
                    return json.dumps(serialize(M.ProtocolErrorMessage(protocol_msg="Message deserialization failed")))

                result = await self.disp.dispatch_message(message)
                return json.dumps(serialize(result))

        self.endpoint = WebsocketRPCEndpoint(AggregatorRpcMethods(self),
                                             on_connect=[self.on_client_connect],  # type: ignore
                                             on_disconnect=[self.on_client_disconnect])  # type: ignore
        self.endpoint.register_route(self.router, path=AGGREGATOR_RPC_WS_PATH)
        self._register_post_route(self.router)

    def _register_post_route(self, router: APIRouter | FastAPI):
        """
        Set up post route to handle RegisterEngineMsg which must be handled before websocket can be established
        """
        @router.post(AGGREGATOR_REST_PATH, response_model_exclude_none=True)
        async def post(request: Request):
            request_json = await request.json()
            try:
                message = deserialize(request_json)
                assert isinstance(message, EM.RegisterEngineMsg), "type must be RegisterEngineMsg"
            except Exception:
                logger.error(f"Message deserialization failed, json:\n{request_json}\n", exc_info=True)
                raise ProtocolException("Deserialization error")

            if self._register_handler is None:
                logger.error("Missing handler for registering engine")
                return M.ProtocolErrorMessage(protocol_msg="Missing handler for registering engine")

            response_message = await self._register_handler(message)
            if response_message is None:
                logger.error(f"Invalid response (None) returned from handler for message type {type(message).__name__}")
                response_message = M.ProtocolErrorMessage(protocol_msg="Invalid response from handler")

            try:
                message_json = serialize(response_message)
            except Exception:
                logger.error(f"Message serialization failed, message type: {type(response_message).__name__}")
                raise ProtocolException("Serialization error")

            return message_json

    async def _on_delayed_client_connect(self, channel: RpcChannel):
        """ We 'delay' our on_connect callback because the WebsocketRPCEndpoint calls on_connect callbacks before it starts
        listening to responses to rpc calls. When we use create_task(), in on_client_connect() below, to call this method,
        we ensure that WebsocketRPCEndpoint starts listening to responses before we call get_engine_id_async() over rpc.
        """
        try:
            response = await channel.other.get_engine_id_async()
        except Exception:
            logger.error("Failed to invoke 'channel.other.get_engine_id_async()'", exc_info=True)
            return

        try:
            assert isinstance(response, RpcResponse)
            engine_id = response.result
            assert isinstance(engine_id, str | None)
            if engine_id is None:
                logger.error("Engine tried connecting with no engine_id available. Closing connection.")
                await channel.close()
                return

            if engine_id in self._engine_id_channel_map:
                logger.warning("Engine tried connecting with engine_id that is already connected. Closing connection.")
                await channel.close()
                return

            logger.info(f"Engine connected: '{engine_id}'")
            self._engine_id_channel_map[engine_id] = channel

            if self._connect_handler is not None:
                await self._connect_handler(engine_id)
        except Exception:
            logger.error("on_delayed_client_connect failed with exception", exc_info=True)
            try:
                await channel.close()
            except Exception:
                logger.warning("Failed to invoke channel.close()", exc_info=True)

    async def on_client_connect(self, channel: RpcChannel):
        channel.default_response_timeout = WEBSOCKET_RPC_TIMEOUT_SECS
        task = asyncio.create_task(self._on_delayed_client_connect(channel))
        self._on_client_connect_tasks.add(task)
        task.add_done_callback(self._on_client_connect_tasks.discard)

    async def on_client_disconnect(self, channel: RpcChannel):
        engine_id: str | None = None
        for key, value in self._engine_id_channel_map.items():
            if value == channel:
                engine_id = key
                break
        if engine_id is not None:
            logger.info(f"Engine '{engine_id}' disconnected")
            del self._engine_id_channel_map[engine_id]
            # Notify aggregator that engine disconnected so that it
            # can save what is necessary and remove engine_data
            if self._disconnect_handler is not None:
                await self._disconnect_handler(engine_id)
        else:
            logger.error("Unknown engine disconnected")

    def has_connected_engine_id(self, engine_id: str) -> bool:
        """ Return a value indicating whether the engine_id is known. Abstract method. """
        return engine_id in self._engine_id_channel_map.keys()

    async def rpc_call(self, engine_id: str, message: M.MessageBase) -> M.MessageBase:
        logger.info(f"Invoke rpc, engine: {engine_id}, msg-type: {type(message).__name__}")
        logger.debug(f"{message}")
        if engine_id not in self._engine_id_channel_map.keys():
            logger.error(f"Cannot invoke rpc call to unknown engine: {engine_id}")
            return M.ErrorMessage(message=f"Cannot invoke rpc call to unknown engine: {engine_id}")

        channel = self._engine_id_channel_map[engine_id]

        try:
            message_json = serialize(message)
        except Exception:
            msg = f"Failed to seralize message: {type(message).__name__}"
            logger.error(msg, exc_info=True)
            return M.ErrorMessage(message=msg)

        try:
            response = await channel.other.dispatch_message_async(message_json=message_json)
            assert isinstance(response, RpcResponse)
            assert isinstance(response.result, str)
        except Exception:
            logger.error("Error during rpc call", exc_info=True)
            await channel.close()
            # this causes connection to fail and then reconnect
            raise ProtocolException("Error in rpc call")

        try:
            result_json = json.loads(response.result)
            result_message = deserialize(result_json)
            return result_message
        except Exception:
            logger.error("Failed to deserialize rpc result", exc_info=True)
            return M.ErrorMessage(message="Failed to deserialize rpc result")

    async def dispatch_message(self, message: EM.EngineMessage) -> M.MessageBase:
        """ Dispatch incoming message to registered handler. """
        logger.debug(f"Incoming message: {message.ident}")

        if message.engine_id == '':
            logger.warning("Failed to handle engine message as its engine_id was the empty string")
            return M.ProtocolErrorMessage(
                protocol_msg="Failed to handle engine message as its engine_id was the empty string")

        message_type = type(message)
        if message_type in self._handlers.keys():
            try:
                return await self._handlers[message_type](message)
            except Exception:
                logger.error(f"Dispatch failed for message type: {message_type}. Handler raised exception.", exc_info=True)
                return M.ProtocolErrorMessage(protocol_msg="Dispatch failed. Handler raised exception.")
        else:
            logger.warning(f"Dispatch failed for message type: {message_type}. No handler registered.")
            return M.ProtocolErrorMessage(protocol_msg="Dispatch failed. No handler registered.")

    def set_message_handler(self, message_type: type[EngineMessageType], handler: AggregatorMessageHandler):
        """ Set handler for message_type. """
        if message_type in self._handlers.keys():
            logger.error(f"Handler for message type {message_type} is already set.")
        self._handlers[message_type] = handler

    def set_register_handler(self, handler: RegisterHandler):
        self._register_handler = handler

    def set_connect_handler(self, handler: EngineConnectHandler):
        self._connect_handler = handler

    def set_disconnect_handler(self, handler: EngineDisconnectHandler):
        self._disconnect_handler = handler

    async def shutdown(self):
        logger.info("shutdown started")
        self._handlers.clear()
        for _, channel in self._engine_id_channel_map.items():
            await channel.close()
        self._disconnect_handler = None
        logger.info("shutdown completed")



"""Added ErrorLog to RecentRun

Revision ID: d517729a9be0
Revises: 5dbc834cf70e
Create Date: 2024-02-12 14:22:09.812491

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'd517729a9be0'
down_revision: Union[str, None] = '5dbc834cf70e'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('RecentRunErrorLogs',
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('error_log', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentRunErrorLogs'))
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('RecentRunErrorLogs')
    # ### end Alembic commands ###



"""
Defines the standard imports for uod modules.
"""
from openpectus.engine.hardware import HardwareLayerBase, Register, RegisterDirection
from openpectus.lang.exec.uod import (
    UnitOperationDefinitionBase, UodCommand, UodBuilder,
    RegexNumber, RegexText, RegexCategorical,
)
from openpectus.protocol.models import PlotConfiguration, SubPlot, PlotAxis, PlotColorRegion
import openpectus.lang.exec.tags_impl as tags
import openpectus.lang.exec.units as units
from openpectus.lang.exec.units import as_float, as_int

__all__ = [
    "HardwareLayerBase", "Register", "RegisterDirection",
    "UnitOperationDefinitionBase", "UodCommand", "UodBuilder",
    "RegexNumber", "RegexText", "RegexCategorical",
    "PlotConfiguration", "SubPlot", "PlotAxis", "PlotColorRegion",
    "tags", "units", "as_float", "as_int"
]




from io import StringIO
from openpectus.lang.grammar.pgrammar import PGrammar
from openpectus.lang.model.pprogram import (
    PNode,
    PProgram,
    PBlank
)


def print_program(program: PProgram, show_line_numbers: bool = False, show_errors: bool = False, show_blanks: bool = False):
    """ Print program to stdout using the provided options. """
    opts = FormattingOptions()
    opts.line_numbers = show_line_numbers
    opts.errors = show_errors
    opts.blanks = show_blanks
    out = PProgramFormatter(opts).format(program)
    print(out)


def print_parsed_program(pcode: str, show_line_numbers: bool = False, show_errors: bool = False, show_blanks: bool = False):
    """ Parse pcode and print program to stdout using the provided options. """
    p = PGrammar()
    p.parse(pcode)
    program = p.build_model()
    return print_program(program, show_line_numbers=show_line_numbers, show_errors=show_errors, show_blanks=show_blanks)


class FormattingOptions():
    def __init__(self) -> None:
        self.indent: int = 4
        self.blanks: bool = False
        self.line_numbers: bool = False
        self.errors: bool = False


class PProgramFormatter():
    def __init__(self, opts: FormattingOptions = FormattingOptions()) -> None:
        self.opts = opts
        self.out = StringIO()

    def format(self, program: PProgram) -> str:
        def print_node(node: PNode, indent: int):
            if not self.opts.blanks and isinstance(node, PBlank):
                return
            self.out.write("".join(" " for _ in range(indent)))
            self.out.write(type(node).__name__)
            if self.opts.errors:
                err = node.errors[0].message if node.has_error() else ""  # type: ignore
                self.out.write(f" | Err: {err}")
            if self.opts.line_numbers:
                self.out.write(f" | Line: {node.line}")

            self.out.write("\n")
            for child in node.get_child_nodes():
                print_node(child, indent + self.opts.indent)

        print_node(program, 0)
        return self.out.getvalue().strip()



import logging

import openpectus.aggregator.deps as agg_deps
import openpectus.aggregator.models as Mdl
import openpectus.aggregator.routers.dto as Dto
from fastapi import APIRouter, Depends, Response, HTTPException
from openpectus.aggregator import command_util
from openpectus.aggregator.aggregator import Aggregator
from openpectus.aggregator.command_examples import examples
from openpectus.aggregator.data import database
from openpectus.aggregator.data.repository import PlotLogRepository, RecentEngineRepository
from openpectus.aggregator.routers.auth import has_access, UserRolesValue, UserNameValue
from starlette.status import HTTP_403_FORBIDDEN, HTTP_404_NOT_FOUND

logger = logging.getLogger(__name__)
router = APIRouter(tags=["process_unit"])


def map_pu(engine_data: Mdl.EngineData) -> Dto.ProcessUnit:
    state = Dto.ProcessUnitState.Ready(state=Dto.ProcessUnitStateEnum.READY)
    if engine_data.system_state is not None and engine_data.system_state.value == Mdl.SystemStateEnum.Running:
        state = Dto.ProcessUnitState.InProgress(
            state=Dto.ProcessUnitStateEnum.IN_PROGRESS,
            progress_pct=0  # TODO: how do we know the progress_pct?
        )
    elif engine_data.has_run() and engine_data.run_data.interrupted_by_error:
        state = Dto.ProcessUnitState.Error(
            state=Dto.ProcessUnitStateEnum.ERROR
        )

    unit = Dto.ProcessUnit(
        id=engine_data.engine_id or "(error)",
        name=f"{engine_data.computer_name} ({engine_data.uod_name})",
        state=state,
        location=engine_data.location,
        runtime_msec=engine_data.runtime.value if (
            engine_data.runtime is not None and isinstance(engine_data.runtime.value, int)
        ) else 0,
        current_user_role=Dto.UserRole.ADMIN,
        uod_author_name=engine_data.uod_author_name,
        uod_author_email=engine_data.uod_author_email
    )
    return unit


def get_registered_engine_data_or_fail(engine_id: str, user_roles: UserRolesValue, agg: Aggregator) -> Mdl.EngineData:
    engine_data = agg.get_registered_engine_data(engine_id)
    if engine_data is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND)
    if not has_access(engine_data, user_roles):
        raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail={'missing_roles': list(engine_data.required_roles)})
    return engine_data


@router.get("/process_unit/{unit_id}", response_model_exclude_none=True)
def get_unit(user_roles: UserRolesValue, unit_id: str, agg: Aggregator = Depends(agg_deps.get_aggregator)) \
        -> Dto.ProcessUnit:
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    return map_pu(engine_data=engine_data)


@router.get("/process_units", response_model_exclude_none=True)
def get_units(user_roles: UserRolesValue, agg: Aggregator = Depends(agg_deps.get_aggregator)) -> list[Dto.ProcessUnit]:
    units: list[Dto.ProcessUnit] = []
    all_engine_data = agg.get_all_registered_engine_data()
    for engine_data in all_engine_data:
        if not has_access(engine_data, user_roles):
            continue
        unit = map_pu(engine_data)
        units.append(unit)
    # append recent engines from the database
    repo = RecentEngineRepository(database.scoped_session())
    recent_engines = repo.get_recent_engines()
    for recent_engine in recent_engines:
        if not has_access(recent_engine, user_roles):
            continue
        if recent_engine.engine_id not in [e.engine_id for e in all_engine_data]:
            unit = Dto.ProcessUnit(
                id=recent_engine.engine_id or "(error)",
                name=recent_engine.name,
                state=Dto.ProcessUnitState.NotOnline(
                    state=Dto.ProcessUnitStateEnum.NOT_ONLINE,
                    last_seen_date=recent_engine.last_update
                ),
                location=recent_engine.location,
                runtime_msec=0,
                current_user_role=Dto.UserRole.ADMIN,
            )
            units.append(unit)

    return units

@router.get("/process_unit/{engine_id}/process_values", response_model_exclude_none=True)
def get_process_values(
        user_roles: UserRolesValue,
        engine_id: str,
        response: Response,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> list[Dto.ProcessValue]:
    response.headers["Cache-Control"] = "no-store"

    engine_data = get_registered_engine_data_or_fail(engine_id, user_roles, agg)
    tags_info = engine_data.tags_info.map
    process_values: list[Dto.ProcessValue] = []
    for reading in engine_data.readings:
        tag_value = tags_info.get(reading.tag_name)
        if tag_value is not None:
            try:
                cmds = command_util.create_reading_commands(tag_value, reading)
                process_values.append(Dto.ProcessValue.create_w_commands(tag_value, cmds))
            except Exception as ex:
                logger.error(f"Error creating commands for process value '{reading.tag_name}': {ex}")
    return process_values


@router.get('/process_unit/{engine_id}/all_process_values', response_model_exclude_none=True)
def get_all_process_values(
        user_roles: UserRolesValue,
        engine_id: str,
        response: Response,
        agg: Aggregator = Depends(agg_deps.get_aggregator)
        ) -> list[Dto.ProcessValue]:
    response.headers["Cache-Control"] = "no-store"
    engine_data = get_registered_engine_data_or_fail(engine_id, user_roles, agg)
    tags_info = engine_data.tags_info.map
    process_values: list[Dto.ProcessValue] = []
    for tag_value in tags_info.values():
        matching_reading = next((r for r in engine_data.readings if r.tag_name == tag_value.name), None)
        if matching_reading is not None:
            try:
                cmds = command_util.create_reading_commands(tag_value, matching_reading)
                process_values.append(Dto.ProcessValue.create_w_commands(tag_value, cmds))
            except Exception as ex:
                logger.error(f"Error creating commands for process value '{matching_reading.tag_name}': {ex}")
        else:
            process_values.append(Dto.ProcessValue.create(tag_value))
    return process_values


@router.post("/process_unit/{unit_id}/execute_command", response_model_exclude_none=True)
async def execute_command(
        user_name: UserNameValue,
        user_roles: UserRolesValue,
        unit_id: str,
        command: Dto.ExecutableCommand,
        agg: Aggregator = Depends(agg_deps.get_aggregator)):
    if __debug__:
        print("ExecutableCommand", command)
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    try:
        msg = command_util.parse_as_message(command, engine_data.readings)
    except Exception:
        logger.error(f"Parse error for command: {command}", exc_info=True)
        return Dto.ServerErrorResponse(message="Message parse error")
    logger.info(f"Sending msg '{str(msg)}' of type {type(msg)} to engine '{unit_id}'")
    try:
        await agg.dispatcher.rpc_call(unit_id, msg)
    except Exception:
        logger.error(f"Rpc call to engine_id '{unit_id}' failed", exc_info=True)
        return Dto.ServerErrorResponse(message="Failed to send message")

    # for now, all users issuing a command become contributors. may nee to filter that somehow
    # and when wo we clear the contributors?
    engine_data.contributors.add(user_name)
    return Dto.ServerSuccessResponse()


@router.get("/process_unit/{unit_id}/process_diagram", response_model_exclude_none=True)
def get_process_diagram(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.ProcessDiagram:
    return Dto.ProcessDiagram(svg="")


@router.get('/process_unit/{unit_id}/command_examples', response_model_exclude_none=True)
def get_command_examples(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> list[Dto.CommandExample]:
    commands: list[Dto.CommandExample] = []
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    commands.append(Dto.CommandExample(name="--- UOD Commands ---", example=""))
    commands.extend(command_util.create_command_examples(engine_data.commands))
    commands.append(Dto.CommandExample(name="--- Example Commands ---", example=""))
    commands.extend(examples)
    return commands


@router.get('/process_unit/{unit_id}/run_log', response_model_exclude_none=True)
def get_run_log(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.RunLog:
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    if not engine_data.has_run():
        return Dto.RunLog.empty()
    else:
        return Dto.RunLog(
            lines=list(map(Dto.RunLogLine.from_model, engine_data.run_data.runlog.lines))
        )


@router.get('/process_unit/{unit_id}/method-and-state', response_model_exclude_none=True)
def get_method_and_state(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.MethodAndState:
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)

    def from_models(method: Mdl.Method, method_state: Mdl.MethodState) -> Dto.MethodAndState:
        return Dto.MethodAndState(
            method=Dto.Method(lines=[Dto.MethodLine(id=line.id, content=line.content) for line in method.lines]),
            state=Dto.MethodState(started_line_ids=[_id for _id in method_state.started_line_ids],
                                  executed_line_ids=[_id for _id in method_state.executed_line_ids],
                                  injected_line_ids=[_id for _id in method_state.injected_line_ids])
        )

    return from_models(engine_data.method, engine_data.method_state)


@router.post('/process_unit/{unit_id}/method', response_model_exclude_none=True)
async def save_method(
        user_name: UserNameValue,
        user_roles: UserRolesValue,
        unit_id: str,
        method_dto: Dto.Method,
        agg: Aggregator = Depends(agg_deps.get_aggregator)):
    _ = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    method_mdl = Mdl.Method(lines=[Mdl.MethodLine(id=line.id, content=line.content) for line in method_dto.lines])

    if not await agg.from_frontend.method_saved(engine_id=unit_id, method=method_mdl, user_name=user_name):
        return Dto.ServerErrorResponse(message="Failed to set method")


@router.get('/process_unit/{unit_id}/plot_configuration', response_model_exclude_none=True)
def get_plot_configuration(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.PlotConfiguration:
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    return Dto.PlotConfiguration.model_validate(
        # assumes Dto.PlotConfiguration and Mdl.PlotConfiguration are identical, change this when they diverge
        engine_data.plot_configuration)


@router.get('/process_unit/{unit_id}/plot_log', response_model_exclude_none=True)
def get_plot_log(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.PlotLog:
    plot_log_repo = PlotLogRepository(database.scoped_session())
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    if engine_data is None or not engine_data.has_run():
        return Dto.PlotLog(entries={})
    plot_log_model = plot_log_repo.get_plot_log(engine_data.run_data.run_id)
    if plot_log_model is None:
        return Dto.PlotLog(entries={})
    # assumes Dto.PlotLog and Mdl.PlotLog are identical, change this when they diverge
    return Dto.PlotLog.model_validate(plot_log_model)


@router.get('/process_unit/{unit_id}/control_state', response_model_exclude_none=True)
def get_control_state(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.ControlState:
    def from_message(state: Mdl.ControlState) -> Dto.ControlState:
        return Dto.ControlState(
            is_running=state.is_running,
            is_holding=state.is_holding,
            is_paused=state.is_paused)

    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    return from_message(engine_data.control_state)


@router.get('/process_unit/{unit_id}/error_log', response_model_exclude_none=True)
def get_error_log(
        user_roles: UserRolesValue,
        unit_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)) -> Dto.AggregatedErrorLog:
    engine_data = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    return Dto.AggregatedErrorLog.from_model(engine_data.error_log)


@router.post('/process_unit/{unit_id}/run_log/force_line/{line_id}', response_model_exclude_none=True)
async def force_run_log_line(
        user_name: UserNameValue,
        user_roles: UserRolesValue,
        unit_id: str,
        line_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)):
    _ = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    if not await agg.from_frontend.request_force(engine_id=unit_id, line_id=line_id, user_name=user_name):
        return Dto.ServerErrorResponse(message="Force request failed")
    return Dto.ServerSuccessResponse(message="Force successfully requested")


@router.post('/process_unit/{unit_id}/run_log/cancel_line/{line_id}', response_model_exclude_none=True)
async def cancel_run_log_line(
        user_name: UserNameValue,
        user_roles: UserRolesValue,
        unit_id: str,
        line_id: str,
        agg: Aggregator = Depends(agg_deps.get_aggregator)):
    _ = get_registered_engine_data_or_fail(unit_id, user_roles, agg)
    if not await agg.from_frontend.request_cancel(engine_id=unit_id, line_id=line_id, user_name=user_name):
        return Dto.ServerErrorResponse(message="Cancel request failed")
    return Dto.ServerSuccessResponse(message="Cancel successfully requested")



@router.get('/process_units/system_state_enum', response_model_exclude_none=True)
def expose_system_state_enum() -> Dto.SystemStateEnum:
    return Mdl.SystemStateEnum.Running



from typing import Iterable, Literal
import uuid
from inspect import cleandoc
from itertools import combinations

from openpectus.engine.models import EntryDataType
from openpectus.lang.exec.tags import Tag, TagCollection
from openpectus.lang.exec.units import get_compatible_unit_names
from openpectus.lang.exec.errors import UodValidationError
import openpectus.protocol.models as Mdl

class ReadingCommand():
    """ Defines commands that are available in the frontend for a given reading/process value. """
    def __init__(self, name: str, command: str) -> None:
        self.command_id: str = str(uuid.uuid4())
        self.name: str = name
        self.command: str = command
        self.choice_names: list[str] = []


class Reading():
    """ Defines a reading that is displayed to the frontend user as a Process Value.

    It is typically defined by matching it with a tag to display that tag's value and with one or
    more commands to allow the user to invoke these commands. The cases where commands are added are
    handled by Reading subclasses.

    execute_command_name is the name of the command to execute when the command is posted from the
    frontend. It defaults to tag_name which is usually fine. It can be set to something else if so
    desired.
    """

    def __init__(self, tag_name: str) -> None:
        if tag_name is None or tag_name == "":
            raise ValueError("tag_name in Reading must not be empty")
        self.discriminator = "reading"
        self.tag_name = tag_name
        self.commands : list[ReadingCommand] = []
        self.tag: Tag | None = None
        self.valid_value_units: list[str] | None = None
        self.entry_data_type: EntryDataType | None = None
        self.command_options: dict[str, str] | None = None

    def match_with_tags(self, tags: TagCollection):
        """ Match the reading with the provided tags. Raises UodValidationError or error. """

        if not tags.has(self.tag_name):
            raise UodValidationError(f"Reading with tag name: {self.tag_name} has no matching tag defined")

        self.tag = tags[self.tag_name]
        self.valid_value_units = get_compatible_unit_names(self.tag.unit)

    def build_commands_list(self):
        """ Override to specify the commands available in the UI for this process value.

        Is invoked after validation and match_with_tags so `tag` and `valid_value_units` have been set.
        """
        pass

    def resolve_entry_type(self):
        """ If entry_data_type is 'auto', resolve its type using the tag's current value.

        If resolved, mutate reading to use the resolved type as entry_data_type.
        """
        if self.entry_data_type == "auto":  # try to resolve the "auto" entry type
            assert self.tag is not None
            value = self.tag.get_value()
            if value is None:
                # Cannot resolve auto type when the tag has no value.
                raise ValueError(f"Cannot resolve entry type 'auto' for tag '{self.tag_name}' because the tag " +
                                 "value is None. Avoid this by setting a tag default value.")
            elif isinstance(value, str):
                self.entry_data_type = "str"
            elif isinstance(value, float):
                self.entry_data_type = "float"
            elif isinstance(value, int):
                self.entry_data_type = "int"

    def as_reading_info(self) -> Mdl.ReadingInfo:
        return Mdl.ReadingInfo(
            discriminator=self.discriminator,
            tag_name=self.tag_name,
            valid_value_units=self.valid_value_units,
            entry_data_type=self.entry_data_type,
            command_options=self.command_options,
            commands=[Mdl.ReadingCommand(
                command_id=c.command_id,
                name=c.name,
                command=c.command,
                choice_names=c.choice_names,
            ) for c in self.commands]
        )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.tag_name=})"

class ReadingWithEntry(Reading):
    def __init__(
            self,
            tag_name: str,
            entry_data_type: EntryDataType = "auto",
            execute_command_name: str | None = None) -> None:
        """ Create a reading with a command that takes an input value from the user.

        Parameters:
            tag_name:
                The tag to display
            entry_data_type: optional, default="auto"
                The data type to display in the UI. Default is auto which uses the type of the tag's value.
            execute_command_name: optional
                The command name. Defaults to the name of the tag.
        """
        super().__init__(tag_name)
        self.execute_command_name = execute_command_name or tag_name
        self.discriminator = "reading_with_entry"
        self.entry_data_type = entry_data_type

    def match_with_tags(self, tags: TagCollection):
        super().match_with_tags(tags)
        assert isinstance(self.tag, Tag)

        if self.entry_data_type == "int":
            if not isinstance(self.tag.value, int):
                raise UodValidationError(
                    f"The process value for tag '{self.tag_name}' specifies the entry_data_type: {self.entry_data_type}. " +
                    f"This conflicts with the tag's current value '{self.tag.value}' which is of type {type(self.tag.value)}"
                )
        elif self.entry_data_type == "float":
            if not isinstance(self.tag.value, float):
                raise UodValidationError(
                    f"The process value for tag '{self.tag_name}' specifies the entry_data_type: {self.entry_data_type}. " +
                    f"This conflicts with the tag's current value '{self.tag.value}' which is of type {type(self.tag.value)}"
                )

    def build_commands_list(self):
        self.commands.append(ReadingCommand(self.tag_name, self.execute_command_name))

class ReadingWithChoice(Reading):
    """ A reading that is displayed with a list of possible commands. """

    def __init__(self, tag_name: str, command_options: dict[str, str] | None = None) -> None:
        """ Create a reading with command choices.

        Parameters:
            tag_name: str
                The tag to display. The tag's choices are used as commands, unless
                command_options are specified.
            command_options: optional
                Dictionary that maps command names to their pcode implementation.
                If not specified, the tag's choices are used as both names and
                pcode implementation. If specified, the tag's choices are not used and the
                tag does not need to have choices defined.
        """
        super().__init__(tag_name)
        self.discriminator = "reading_with_choice"
        self.command_options = command_options

    def match_with_tags(self, tags: TagCollection):
        """ Internal, used in validation."""
        super().match_with_tags(tags)

        assert self.tag is not None
        if self.tag.choices is None or len(self.tag.choices) == 0:
            if self.command_options is None or len(self.command_options) == 0:
                raise UodValidationError(
                    "ReadingWithChoice requires either a tag with choices or non-empty command_options, " +
                    f"Reading Tag: {self.tag_name}")

        if self.command_options is None or len(self.command_options.keys()) == 0:
            self.command_options = {}
            assert self.tag.choices is not None
            for choice in self.tag.choices:
                self.command_options[choice] = choice

    def build_commands_list(self):
        """ Internal, used during validation. """
        assert self.command_options is not None
        choices = list(self.command_options.keys())
        command = ReadingCommand(self.tag_name, "(empty)")
        command.choice_names = choices
        self.commands.append(command)


class ReadingCollection(Iterable[Reading]):
    def __init__(self) -> None:
        super().__init__()
        self._readings: list[Reading] = []

    def add(self, reading: Reading):
        self._readings.append(reading)

    def __iter__(self):
        yield from self._readings

    def __len__(self) -> int:
        return len(self._readings)


class UodCommandDescription:
    """ Describes a command and its usage. """
    def __init__(self, name: str) -> None:
        self.name: str = name
        """ Name of the command as used in pcode. """
        self.docstring: str = ""
        """ Description of the command's purpose, provided by uod author as docstring
        on the exec function. """
        self.argument_type: Literal["float", "unknown", "none"] = "unknown"
        """ the type of the argument in case the commands takes a single argument """
        self.argument_valid_units: list[str] = []
        """ Valid argument units if applicable. """
        self.argument_exclusive_options: list[str] = []
        """ Exclusive options if applicable. """
        self.argument_additive_options: list[str] = []
        """ Additive options if applicable. """

    def as_command_info(self) -> Mdl.CommandInfo:
        return Mdl.CommandInfo(
            name=self.name,
            docstring=self.docstring,
        )

    def set_docstring(self, docstring: str | None):
        self.docstring = cleandoc(docstring) if docstring is not None else ""

    def get_docstring_pcode(self) -> str:
        return self.docstring

    def generate_pcode_examples(self) -> list[str]:
        """ For commands with regex arg parser, generate pcode examples to test the command
        using combinations of argument values and units.
        """
        examples: list[str] = []
        if self.argument_type == "float":
            args = [0.5, 10, 32.6]
            if len(self.argument_valid_units) == 0:
                for arg in args:
                    examples.append(f"{self.name}: {arg}")
            else:
                for unit in self.argument_valid_units:
                    examples.append(f"{self.name}: {args[0]} {unit}")
        else:
            # Create examples for all exclusive options
            # If there are no exclusive options then the list is ['']
            # which we do not want to test.
            for arg in self.argument_exclusive_options:
                if arg:
                    examples.append(f"{self.name}: {arg}")
            # Create examples of all combinations of all possible lengths of additive options
            # If there are no additive options then the list is ['']
            # which we do not want to test.
            if len(self.argument_additive_options):
                for r in range(1, len(self.argument_additive_options)+1):
                    for args in combinations(self.argument_additive_options, r):
                        arg = "+".join(args)
                        if arg:
                            examples.append(f"{self.name}: {arg}")
        return examples

    def __str__(self):
        return f"{self.__class__.__name__}({self.name=})"




class BaseUnitProvider:
    """ Manages a mapping between units that can be used in the Base instruction and
    the tags that provides the value in non-block and block scopes. """
    def __init__(self) -> None:
        self.map: dict[str, tuple[str, str]] = {}

    def has(self, unit: str) -> bool:
        return unit in self.map.keys()

    def set(self, unit: str, main_tag_name: str, block_tag_name: str):
        self.map[unit] = (main_tag_name, block_tag_name)

    def get_units(self) -> list[str]:
        return list(self.map.keys())

    def get_tags(self, unit: str) -> tuple[str, str]:
        if unit not in self.map.keys():
            raise ValueError(f"Unit '{unit}' is not registered")
        return self.map[unit]



from fastapi import APIRouter

from openpectus import __version__, build_number, build_info
from openpectus.aggregator.routers.dto import BuildInfo

router = APIRouter(tags=["version"])

@router.get("/version", response_model_exclude_none=True)
def get_version():
    return __version__

@router.get("/build_number", response_model_exclude_none=True)
def get_build_number():
    return build_number

@router.get("/api/build_info", response_model_exclude_none=True)
def get_build_info() -> BuildInfo:
    return build_info



import logging
import unittest
from typing import Any, Iterable
from openpectus.engine.hardware import (
    HardwareLayerBase,
    Register,
    RegisterDirection,
    HardwareLayerException,
)
from openpectus.engine.composite_hardware import Composite_Hardware

logging.basicConfig(format=" %(name)s :: %(levelname)-8s :: %(message)s")
logger = logging.getLogger("openpectus.engine.composite_hardware.Composite_Hardware")
logger.setLevel(logging.DEBUG)


global_list = []


class Test_Hardware(HardwareLayerBase):
    __test__ = False

    """ Test hardware class which logs all interactions to a global list """
    def __init__(self, identity: int) -> None:
        super().__init__()
        self.identity = identity

    def validate_offline(self):
        global global_list
        for r in self.registers.values():
            global_list.append(("VALIDATE_OFFLINE", str(self), str(r),))

    def validate_online(self):
        global global_list
        for r in self.registers.values():
            global_list.append(("VALIDATE_ONLINE", str(self), str(r),))

    def read(self, r: Register) -> Any:
        global global_list
        global_list.append(("READ", str(self), str(r),))
        return r.options["value"]

    def read_batch(self, registers: Iterable[Register]) -> list[Any]:
        global global_list
        global_list.append(("READ", str(self), str(registers),))
        return [r.options["value"] for r in registers]

    def write(self, value: Any, r: Register):
        global global_list
        global_list.append(("WRITE", str(self), str(r), value,))

    def write_batch(self, values: Iterable[Any], registers: Iterable[Register]):
        global global_list
        global_list.append(("WRITE", str(self), str(registers), str(values),))

    def connect(self):
        global global_list
        global_list.append(("CONNECT", str(self),))

    def disconnect(self):
        global global_list
        global_list.append(("DISCONNECT", str(self),))

    def __str__(self):
        return f"Test_Hardware({self.identity})"

    def __repr__(self):
        return str(self)


def setup_composite_hardware(i0=0):
    hardwares = [Test_Hardware(i0+i) for i in range(3)]

    composite_hardware = Composite_Hardware()
    registers = dict()
    for j, hardware in enumerate(hardwares):
        for i in range(2):
            registers[f"Reg {j}_{i}"] = Register(f"Reg {j}_{i}", RegisterDirection.Both, hardware=hardware, value=i+j*10)

    composite_hardware._registers = registers
    return composite_hardware


class TestCompositeHardware(unittest.TestCase):
    def test_connect(self):
        global global_list
        global_list = []
        composite_hardware = setup_composite_hardware()

        with composite_hardware:
            pass

        self.assertListEqual(global_list, [
            ("CONNECT", "Test_Hardware(0)"),
            ("CONNECT", "Test_Hardware(1)"),
            ("CONNECT", "Test_Hardware(2)"),
            ("DISCONNECT", "Test_Hardware(0)"),
            ("DISCONNECT", "Test_Hardware(1)"),
            ("DISCONNECT", "Test_Hardware(2)")
        ])

    def test_read_register(self):
        global global_list
        global_list = []
        composite_hardware = setup_composite_hardware()
        registers = composite_hardware.registers

        with composite_hardware:
            read = composite_hardware.read(registers["Reg 1_0"])

        self.assertEqual(read, 10)
        self.assertListEqual(global_list, [
            ("CONNECT", "Test_Hardware(0)"),
            ("CONNECT", "Test_Hardware(1)"),
            ("CONNECT", "Test_Hardware(2)"),
            ("READ", "Test_Hardware(1)", "Register(name=Reg 1_0,direction=RegisterDirection.Both)"),
            ("DISCONNECT", "Test_Hardware(0)"),
            ("DISCONNECT", "Test_Hardware(1)"),
            ("DISCONNECT", "Test_Hardware(2)")
        ])

    def test_write_register(self):
        global global_list
        global_list = []
        composite_hardware = setup_composite_hardware()
        registers = composite_hardware.registers

        with composite_hardware:
            composite_hardware.write(10.1, registers["Reg 2_1"])

        self.assertListEqual(global_list, [
            ("CONNECT", "Test_Hardware(0)"),
            ("CONNECT", "Test_Hardware(1)"),
            ("CONNECT", "Test_Hardware(2)"),
            ("WRITE", "Test_Hardware(2)", "Register(name=Reg 2_1,direction=RegisterDirection.Both)", 10.1),
            ("DISCONNECT", "Test_Hardware(0)"),
            ("DISCONNECT", "Test_Hardware(1)"),
            ("DISCONNECT", "Test_Hardware(2)")
        ])

    def test_read_multiple_registers(self):
        global global_list
        global_list = []
        composite_hardware = setup_composite_hardware()
        registers = composite_hardware.registers

        with composite_hardware:
            read = composite_hardware.read_batch([registers[name] for name in ["Reg 2_1",
                                                                               "Reg 0_0",
                                                                               "Reg 0_1",
                                                                               "Reg 2_0",
                                                                               "Reg 1_1"]])

        self.assertListEqual(read, [21, 0, 1, 20, 11])
        self.assertListEqual(global_list, [
            ("CONNECT", "Test_Hardware(0)"),
            ("CONNECT", "Test_Hardware(1)"),
            ("CONNECT", "Test_Hardware(2)"),
            ("READ", "Test_Hardware(2)",
                "[Register(name=Reg 2_1,direction=RegisterDirection.Both), " +
                "Register(name=Reg 2_0,direction=RegisterDirection.Both)]"),
            ("READ", "Test_Hardware(0)",
                "[Register(name=Reg 0_0,direction=RegisterDirection.Both), " +
                "Register(name=Reg 0_1,direction=RegisterDirection.Both)]"),
            ("READ", "Test_Hardware(1)",
                "[Register(name=Reg 1_1,direction=RegisterDirection.Both)]"),
            ("DISCONNECT", "Test_Hardware(0)"),
            ("DISCONNECT", "Test_Hardware(1)"),
            ("DISCONNECT", "Test_Hardware(2)")
        ])

    def test_write_multiple_registers(self):
        global global_list
        global_list = []
        composite_hardware = setup_composite_hardware()
        registers = composite_hardware.registers

        with composite_hardware:
            regs = [registers[name] for name in ["Reg 2_1",
                                                 "Reg 0_0",
                                                 "Reg 0_1",
                                                 "Reg 2_0",
                                                 "Reg 1_1"]]
            vals = [1, 2, 3, 4, 5,]
            composite_hardware.write_batch(vals, regs)

        self.assertListEqual(global_list, [
            ("CONNECT", "Test_Hardware(0)"),
            ("CONNECT", "Test_Hardware(1)"),
            ("CONNECT", "Test_Hardware(2)"),
            ("WRITE", "Test_Hardware(2)",
                "[Register(name=Reg 2_1,direction=RegisterDirection.Both), " +
                "Register(name=Reg 2_0,direction=RegisterDirection.Both)]", "[1, 4]"),
            ("WRITE", "Test_Hardware(0)",
                "[Register(name=Reg 0_0,direction=RegisterDirection.Both), " +
                "Register(name=Reg 0_1,direction=RegisterDirection.Both)]", "[2, 3]"),
            ("WRITE", "Test_Hardware(1)",
                "[Register(name=Reg 1_1,direction=RegisterDirection.Both)]", "[5]"),
            ("DISCONNECT", "Test_Hardware(0)"),
            ("DISCONNECT", "Test_Hardware(1)"),
            ("DISCONNECT", "Test_Hardware(2)")
        ])

    def test_missing_hardware_layer(self):
        global global_list
        global_list = []
        composite_hardware = setup_composite_hardware()
        del composite_hardware.registers["Reg 0_1"].options["hardware"]

        with self.assertRaises(HardwareLayerException) as context:
            composite_hardware.validate_offline()
        self.assertEqual(
            "Hardware layer undefined for register Register(name=Reg 0_1,direction=RegisterDirection.Both).",
            str(context.exception)
        )

    def test_hardware_layer_is_itself_composite(self):
        global global_list
        global_list = []
        composite_hardware_A = setup_composite_hardware()
        composite_hardware_B = setup_composite_hardware(i0=10)
        composite_hardware_A.registers["Reg 0_1"].options["hardware"] = composite_hardware_B

        with self.assertRaises(HardwareLayerException) as context:
            composite_hardware_A.validate_offline()
        self.assertEqual(
            (
                "Hardware layer Composite_Hardware(hardwares=[Test_Hardware(10), "
                "Test_Hardware(11), Test_Hardware(12)]) specified for register "
                "Register(name=Reg 0_1,direction=RegisterDirection.Both) is itself a composite hardware layer. "
                "This is not allowed."
            ),
            str(context.exception)
        )


if __name__ == "__main__":
    unittest.main()
